<?php

namespace Drupal\aws_cloud\Controller\Ec2;

use Drupal\aws_cloud\Entity\Ec2\ElasticIp;
use Drupal\aws_cloud\Entity\Ec2\InstanceInterface;
use Drupal\aws_cloud\Entity\Ec2\NetworkInterface;
use Drupal\aws_cloud\Service\AwsCloud\AwsCloudOperationsServiceInterface;
use Drupal\aws_cloud\Service\CloudWatch\CloudWatchServiceInterface;
use Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface;
use Drupal\aws_cloud\Traits\AwsCloudEntityCheckTrait;
use Drupal\cloud\Entity\CloudConfig;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudService;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\cloud\Traits\CloudResourceTrait;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormState;
use Drupal\Core\Link;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\openstack\Entity\IntermediateFormState;
use Drupal\views\Views;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Controller responsible for "update" URLs.
 *
 * This class is mainly responsible for
 * updating the AWS entities from URLs.
 */
class ApiController extends ControllerBase implements ApiControllerInterface {

  use AwsCloudEntityCheckTrait;
  use CloudContentEntityTrait;
  use CloudResourceTrait;

  /**
   * The AWS Cloud Operations service.
   *
   * @var \Drupal\aws_cloud\Service\AwsCloud\AwsCloudOperationsServiceInterface
   */
  protected $awsCloudOperationsService;

  /**
   * The AWS Cloud EC2 service.
   *
   * @var \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface
   */
  private $ec2Service;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * Request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  private $requestStack;

  /**
   * Renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  private $renderer;

  /**
   * The AWS Cloud CloudWatch service.
   *
   * @var \Drupal\aws_cloud\Service\CloudWatch\CloudWatchServiceInterface
   */
  private $cloudWatchService;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * Cloud service interface.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  protected $cloudService;

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The configuration data.
   *
   * @var array
   */
  protected $configuration;

  /**
   * ApiController constructor.
   *
   * @param \Drupal\aws_cloud\Service\AwsCloud\AwsCloudOperationsServiceInterface $aws_cloud_operations_service
   *   The AWS Cloud Operations service.
   * @param \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface $ec2_service
   *   Object for interfacing with AWS API.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   Messenger Object.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   Request stack object.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer service.
   * @param \Drupal\aws_cloud\Service\CloudWatch\CloudWatchServiceInterface $cloud_watch_service
   *   The AWS Cloud CloudWatch service.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   Cloud service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   */
  public function __construct(
    AwsCloudOperationsServiceInterface $aws_cloud_operations_service,
    Ec2ServiceInterface $ec2_service,
    Messenger $messenger,
    RequestStack $request_stack,
    RendererInterface $renderer,
    CloudWatchServiceInterface $cloud_watch_service,
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    CloudServiceInterface $cloud_service,
    EntityTypeManagerInterface $entity_type_manager,
    AccountInterface $current_user
  ) {

    $this->awsCloudOperationsService = $aws_cloud_operations_service;
    $this->ec2Service = $ec2_service;
    $this->messenger = $messenger;
    $this->requestStack = $request_stack;
    $this->renderer = $renderer;
    $this->cloudWatchService = $cloud_watch_service;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->cloudService = $cloud_service;
    $this->entityTypeManager = $entity_type_manager;
    $this->currentUser = $current_user;
    $this->configuration = ['cloud_context' => ''];
  }

  /**
   * Dependency Injection.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   Instance of ContainerInterface.
   *
   * @return ApiController
   *   return created object.
   */
  public static function create(ContainerInterface $container): ApiController {
    return new static(
      $container->get('aws_cloud.operations'),
      $container->get('aws_cloud.ec2'),
      $container->get('messenger'),
      $container->get('request_stack'),
      $container->get('renderer'),
      $container->get('aws_cloud.cloud_watch'),
      $container->get('plugin.manager.cloud_config_plugin'),
      $container->get('cloud'),
      $container->get('entity_type.manager'),
      $container->get('current_user')
    );
  }

  /**
   * {@inheritdoc}
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirects to the cloud_config collection.
   */
  public function updateAll(): RedirectResponse {
    $regions = $this->requestStack->getCurrentRequest()->query->get('regions');
    if ($regions === NULL) {
      $this->messageUser($this->t('No region specified'), 'error');
    }
    else {
      $regions_array = explode(',', $regions);

      foreach ($regions_array ?: [] as $region) {
        $entity = $this->entityTypeManager()->getStorage('cloud_config')
          ->loadByProperties(
            [
              'cloud_context' => $region,
            ]);
        if ($entity) {
          aws_cloud_update_ec2_resources(array_shift($entity));
        }
      }

      $this->messageUser($this->t('Creating cloud service provider was performed successfully.'));
      drupal_flush_all_caches();
    }
    return $this->redirect('entity.cloud_config.collection');
  }

  /**
   * Update message after refresh resources.
   *
   * @param string $resource_type
   *   Entity type name.
   * @param bool $updated
   *   Updated statuses.
   */
  protected function updateMessages(string $resource_type, bool $updated): void {
    $labels = $this->getDisplayLabels($resource_type);
    if ($updated) {
      $this->messageUser($this->t('Updated @resource_name.', ['@resource_name' => $labels['plural']]));
    }
    else {
      $this->messageUser($this->t('Unable to update @resource_name.', ['@resource_name' => $labels['plural']]), 'error');
    }
    $this->cloudService->invalidateCacheTags();
  }

  /**
   * Update all instances in particular cloud region.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to instance list of particular cloud region.
   */
  public function updateInstanceList($cloud_context): RedirectResponse {
    $this->ec2Service->setCloudContext($cloud_context);
    $updated = $this->ec2Service->updateInstances();

    $this->updateMessages('aws_cloud_instance', $updated);

    return $this->redirect('view.aws_cloud_instance.list', [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Update all instances of all cloud region.
   */
  public function updateAllInstanceList(): RedirectResponse {
    return $this->ec2Service->updateAllResourceList('aws_cloud_instance');
  }

  /**
   * Update all images in particular cloud region.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to all image list.
   */
  public function updateImageList($cloud_context): RedirectResponse {
    $cloud_config_entities = $this->entityTypeManager()->getStorage('cloud_config')->loadByProperties(
      ['cloud_context' => [$cloud_context]]
    );

    /** @var \Drupal\cloud\Entity\CloudConfig $cloud_config */
    $cloud_config = !empty($cloud_config_entities)
      ? reset($cloud_config_entities)
      : CloudConfig::create([
        'cloud_context' => $cloud_context,
      ]);

    /** @var \Drupal\aws_cloud\Service\Ec2\Ec2Service */
    $ec2Service = $this->ec2Service;
    $account_id = $ec2Service->getAccountId($cloud_config_entities);

    if ($account_id) {
      $this->ec2Service->setCloudContext($cloud_context);
      $updated = $this->ec2Service->updateImages([
        'Owners' => [
          $account_id,
        ],
      ], TRUE);

      $this->updateMessages('aws_cloud_image', $updated);
    }
    else {
      $message = $this->t('AWS User ID is not specified.');
      $account = $this->currentUser();
      if ($account->hasPermission('edit cloud service providers')) {
        $message = Link::createFromRoute($message, 'entity.cloud_config.edit_form', ['cloud_config' => $cloud_config->id()])->toString();
      }
      $this->messageUser($message, 'error');
    }

    return $this->redirect('view.aws_cloud_image.list', [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Update all images of all cloud region.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to all image list.
   */
  public function updateAllImageList(): RedirectResponse {
    return $this->ec2Service->updateAllResourceList('aws_cloud_image');
  }

  /**
   * Update all security groups in particular cloud region.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to all security group list.
   */
  public function updateSecurityGroupList($cloud_context): RedirectResponse {
    $this->ec2Service->setCloudContext($cloud_context);
    $updated = $this->ec2Service->updateSecurityGroups();

    $this->updateMessages('aws_cloud_security_group', $updated);

    return $this->redirect('view.aws_cloud_security_group.list', [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Update all security groups of all cloud region.
   */
  public function updateAllSecurityGroupList(): RedirectResponse {
    return $this->ec2Service->updateAllResourceList('aws_cloud_security_group');
  }

  /**
   * Update all network interfaces in particular cloud region.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to network interface list of particular cloud region.
   */
  public function updateNetworkInterfaceList($cloud_context): RedirectResponse {
    $this->ec2Service->setCloudContext($cloud_context);
    $updated = $this->ec2Service->updateNetworkInterfaces();

    $this->updateMessages('aws_cloud_network_interface', $updated);

    return $this->redirect('view.aws_cloud_network_interface.list', [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Update all network interfaces of all cloud region.
   */
  public function updateAllNetworkInterfaceList(): RedirectResponse {
    return $this->ec2Service->updateAllResourceList('aws_cloud_network_interface');
  }

  /**
   * Update all elastic ips in particular cloud region.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to elastic ip list of particular cloud region.
   */
  public function updateElasticIpList($cloud_context): RedirectResponse {

    $this->ec2Service->setCloudContext($cloud_context);
    $updated = $this->ec2Service->updateElasticIps();

    if ($updated === TRUE) {
      // Also update network interfaces.
      $updated = $this->ec2Service->updateNetworkInterfaces();
      if ($updated !== FALSE) {
        $this->messageUser($this->t('Updated Elastic IPs and network interfaces.'));
      }
      else {
        $this->messageUser(
          $this->t('Unable to update network interfaces while updating Elastic IPs.'),
          'error');
      }
      $this->cloudService->invalidateCacheTags();
    }
    else {
      $this->messageUser($this->t('Unable to update Elastic IPs.'), 'error');
    }

    return $this->redirect('view.aws_cloud_elastic_ip.list', [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Update all elastic ips of all cloud region.
   */
  public function updateAllElasticIpList(): RedirectResponse {
    return $this->ec2Service->updateAllResourceList('aws_cloud_elastic_ip');
  }

  /**
   * Update all key pairs in particular cloud region.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to keypair list of particular cloud region.
   */
  public function updateKeyPairList($cloud_context): RedirectResponse {

    $this->ec2Service->setCloudContext($cloud_context);
    $updated = $this->ec2Service->updateKeyPairs();

    $this->updateMessages('aws_cloud_key_pair', $updated);
    return $this->redirect('view.aws_cloud_key_pair.list', [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Update all key pairs of all cloud region.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to all instance list.
   */
  public function updateAllKeyPairList(): RedirectResponse {
    return $this->ec2Service->updateAllResourceList('aws_cloud_key_pair');
  }

  /**
   * Update all volumes in particular cloud region.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to volume list of particular cloud region.
   */
  public function updateVolumeList($cloud_context): RedirectResponse {
    $this->ec2Service->setCloudContext($cloud_context);
    $updated = $this->ec2Service->updateVolumes();

    $this->updateMessages('aws_cloud_volume', $updated);

    return $this->redirect('view.aws_cloud_volume.list', [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Update all volumes of all cloud region.
   */
  public function updateAllVolumeList(): RedirectResponse {
    return $this->ec2Service->updateAllResourceList('aws_cloud_volume');
  }

  /**
   * Update all snapshots in particular cloud region.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to snapshot list of particular cloud region.
   */
  public function updateSnapshotList($cloud_context): RedirectResponse {

    $this->ec2Service->setCloudContext($cloud_context);
    $updated = $this->ec2Service->updateSnapshots();

    $this->updateMessages('aws_cloud_snapshot', $updated);

    return $this->redirect('view.aws_cloud_snapshot.list', [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Update all snapshots of all cloud region.
   */
  public function updateAllSnapshotList(): RedirectResponse {
    return $this->ec2Service->updateAllResourceList('aws_cloud_snapshot');
  }

  /**
   * {@inheritdoc}
   */
  public function listInstanceCallback(): Response {
    return $this->getViewResponse('aws_cloud_instance');
  }

  /**
   * {@inheritdoc}
   */
  public function listImageCallback(): Response {
    return $this->getViewResponse('aws_cloud_image');
  }

  /**
   * {@inheritdoc}
   */
  public function listSnapshotCallback(): Response {
    return $this->getViewResponse('aws_cloud_snapshot');
  }

  /**
   * {@inheritdoc}
   */
  public function listVolumeCallback(): Response {
    return $this->getViewResponse('aws_cloud_volume');
  }

  /**
   * {@inheritdoc}
   */
  public function searchImages($cloud_context): Response {
    $this->ec2Service->setCloudContext($cloud_context);
    $name = $this->requestStack->getCurrentRequest()->query->get('q');
    $result = $this->ec2Service->describeImages([
      'Filters' => [
        [
          'Name' => 'name',
          'Values' => [$name],
        ],
        [
          'Name' => 'state',
          'Values' => ['available'],
        ],
      ],
    ]);

    $limit = 50;
    $count = 0;
    $images = [];
    foreach ($result['Images'] ?: [] as $image) {
      $images[] = [
        'name' => $image['Name'],
        'id' => $image['Name'],
        'created' => strtotime($image['CreationDate']),
      ];
      if ($count++ >= $limit) {
        break;
      }
    }

    // Sort by latest.
    usort($images, static function ($a, $b) {
      return ($a['created'] < $b['created']) ? 1 : -1;
    });

    try {
      return new Response(json_encode($images));
    }
    catch (\JsonException $e) {
      $this->handleException($e);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getInstanceMetrics($cloud_context, InstanceInterface $aws_cloud_instance): JsonResponse {
    $this->cloudWatchService->setCloudContext($cloud_context);
    $metric_names = [
      'cpu' => 'CPUUtilization',
      'network_in' => 'NetworkIn',
      'network_out' => 'NetworkOut',
      'disk_read' => 'DiskReadBytes',
      'disk_write' => 'DiskWriteBytes',
      'disk_read_operation' => 'DiskReadOps',
      'disk_write_operation' => 'DiskWriteOps',
    ];
    $queries = [];
    foreach ($metric_names as $key => $name) {
      $queries[] = [
        'Id' => $key,
        'MetricStat' => [
          'Metric' => [
            'Namespace' => 'AWS/EC2',
            'MetricName' => $name,
            'Dimensions' => [
              [
                'Name' => 'InstanceId',
                'Value' => $aws_cloud_instance->getInstanceId(),
              ],
            ],
          ],
          'Period' => 300,
          'Stat' => 'Average',
        ],
      ];
    }

    $result = $this->cloudWatchService->getMetricData([
      'StartTime' => strtotime('-1 days'),
      'EndTime' => strtotime('now'),
      'MetricDataQueries' => $queries,
    ]);

    $data = [];
    foreach (array_keys($metric_names) as $index => $key) {
      $timestamps = [];
      foreach ($result['MetricDataResults'][$index]['Timestamps'] as $timestamp) {
        $timestamps[] = $timestamp->__toString();
      }

      $data[$key] = [
        'timestamps' => $timestamps,
        'values' => $result['MetricDataResults'][$index]['Values'],
      ];

      if ($key === 'network_in' || $key === 'network_out') {
        // Convert Byte to MB.
        $data[$key]['values'] = array_map(function ($value) {
          return $value / 1024 / 1024;
        }, $data[$key]['values']);
      }
    }

    return new JsonResponse($data);
  }

  /**
   * Helper method to get views output.
   *
   * @param string $view_id
   *   The ID of list view.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The response of list view.
   */
  private function getViewResponse(string $view_id): Response {
    $view = Views::getView($view_id);

    // Set the display machine name.
    $view->setDisplay('list');

    // Render the view as html, and return it as a response object.
    $build = $view->executeDisplay();
    return new Response($this->renderer->render($build));
  }

  /**
   * Helper method to add messages for the end user.
   *
   * @param string $message
   *   The message.
   * @param string $type
   *   The message type: error or message.
   */
  private function messageUser(string $message, string $type = 'message'): void {
    switch ($type) {
      case 'error':
        $this->messenger->addError($message);
        break;

      case 'message':
        $this->messenger->addStatus($message);
        break;

      default:
        break;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityCount(string $cloud_context, string $entity_type_id): JsonResponse {
    $params = !empty($cloud_context)
      ? ['cloud_context' => $cloud_context] : [];
    if (!empty($cloud_context)) {
      $this->configuration['cloud_context'] = $cloud_context;
    }
    $count = $this->getResourceCount(
      $entity_type_id,
      // Use a static trait method through CloudService.
      'list ' . CloudService::convertUnderscoreToWhitespace($entity_type_id),
      $params
    );

    return new JsonResponse(['count' => $count]);
  }

  /**
   * {@inheritdoc}
   */
  public function getConsoleOutput(string $cloud_context, string $entity_id): JsonResponse {
    // Get entity data.
    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity */
    $entity = $this->entityTypeManager
      ->getStorage('aws_cloud_instance')
      ->load($entity_id);
    $entity = $this->updateEntitySubmitForm(new FormState(), $entity)
      ? $entity : NULL;

    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The instance has already been deleted.',
      ], 404);
    }

    // Execute AWS API.
    $result = $this->awsCloudOperationsService->getConsoleOutput($entity);

    return !empty($result)
      ? new JsonResponse([
        'result' => 'OK',
        'log' => $result,
      ], 200)
      : new JsonResponse([
        'result' => 'NG',
        'reason' => 'Internal Server Error',
      ], 500);
  }

  /**
   * {@inheritdoc}
   */
  public function refreshEntity(string $cloud_context, string $entity_type_id, string $entity_id): JsonResponse {
    $entity = $this->entityTypeManager
      ->getStorage($entity_type_id)
      ->load($entity_id);

    $this->ec2Service->setCloudContext($cloud_context);
    $this->ec2Service->updateSingleEntity($entity, FALSE, TRUE);

    return new JsonResponse(['result' => 'OK']);
  }

  /**
   * {@inheritdoc}
   */
  public function operateEntity(Request $request, string $cloud_context, string $entity_type_id, string $entity_id, string $command): JsonResponse {
    // Create an instance of the entity.
    // The process splits between adding a new entity
    // and when reading an existing entity.
    $entity = $command !== 'create'
      ? $this->entityTypeManager
        ->getStorage($entity_type_id)
        ->load($entity_id)
      : $this->entityTypeManager
        ->getStorage($entity_type_id)
        ->create([
          'cloud_context' => $cloud_context,
        ]);
    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The instance has already been deleted.',
      ], 404);
    }

    // Depending on the content of the process,
    // branching is performed with a switch statement.
    $form_state = new IntermediateFormState();
    $method_name = '';
    try {
      switch ($command . '_' . $entity_type_id) {
        case 'associate_elastic_ip_aws_cloud_instance':
          $form_state->set('allocation_id', $request->get('allocation_id', ''));
          $form_state->set('network_interface_id', $request->get('network_interface_id', ''));

          $method_name = 'associateElasticIpForInstance';
          break;

        case 'create_image_aws_cloud_instance':
          $form_state->set('image_name', $request->get('image_name', ''));
          $form_state->set('no_reboot', $request->get('no_reboot', 'false') === 'true');

          $method_name = 'createImageFromInstance';
          break;

        case 'delete_aws_cloud_instance':
          $method_name = 'deleteInstance';
          break;

        case 'edit_aws_cloud_instance':
          /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity */
          $entity->setName($request->get('name', ''));
          $entity->setIamRole($request->get('iam_role', ''));
          $form_state->set('security_groups', json_decode($request->get('security_groups', '[]'), TRUE));
          $form_state->set('tags', json_decode($request->get('tags', '[]'), TRUE));
          $form_state->set('termination_timestamp', $request->get('termination_timestamp', NULL));
          $entity->setUserData($request->get('user_data', NULL));
          $entity->setTerminationProtection(json_decode($request->get('termination_protection', '"false"'), TRUE));
          $form_state->set('user_data_base64_encoded', json_decode($request->get('user_data_base64_encoded', '"false"'), TRUE));
          $entity->setSchedule($request->get('schedule', ''));
          $form_state->set('add_new_elastic_ip', $request->get('add_new_elastic_ip', ''));
          $form_state->set('current_allocation_id', $request->get('current_allocation_id', ''));
          $form_state->set('current_association_id', $request->get('current_association_id', ''));

          $method_name = 'editInstance';
          break;

        case 'reboot_aws_cloud_instance':
          $method_name = 'rebootInstance';
          break;

        case 'start_aws_cloud_instance':
          $method_name = 'startInstance';
          break;

        case 'stop_aws_cloud_instance':
          $method_name = 'stopInstance';
          break;
      }
    }
    catch (\JsonException $e) {
      $this->handleException($e);
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'Internal Server Error',
      ], 500);
    }

    // Execute the process.
    $result = NULL;
    if (method_exists($this->awsCloudOperationsService, $method_name)) {
      $form = [];
      $result = $this->awsCloudOperationsService->$method_name(
        $entity,
        $form,
        $form_state
      );
    }

    return !empty($result)
      ? new JsonResponse([
        'result' => 'OK',
      ], 200)
      : new JsonResponse([
        'result' => 'NG',
        'reason' => 'Internal Server Error',
      ], 500);
  }

  /**
   * {@inheritdoc}
   */
  public function getInstanceTypes(string $cloud_context): JsonResponse {
    $instance_type_dict = aws_cloud_get_instance_types($cloud_context);
    $output = [];
    foreach (array_keys($instance_type_dict) as $key) {
      $output[] = [
        // Note: The dictionary type is expressed as Key-Value,
        // and the <option> tag in HTML is expressed by the value attribute
        // and children (labels).
        'value' => $key,
        // Use $key instead of $value because $value in return value of
        // array type of aws_cloud_get_instance_types() is difficult to
        // use as a label.
        'label' => $key,
      ];
    }

    return new JsonResponse($output, 200);
  }

  /**
   * {@inheritdoc}
   */
  public function getIamRoleOptions(string $cloud_context): JsonResponse {
    $iam_role_dict = aws_cloud_get_iam_roles($cloud_context);
    $output = [];
    foreach (array_values($iam_role_dict) as $value) {
      $output[] = [
        'value' => $value,
        'label' => $value,
      ];
    }

    return new JsonResponse($output, 200);
  }

  /**
   * {@inheritdoc}
   */
  public function getSecurityGroupsOptions(string $entity_id): JsonResponse {
    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
    $entity = $this->entityTypeManager
      ->getStorage('aws_cloud_instance')
      ->load($entity_id);
    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The instance has already been deleted.',
      ], 404);
    }

    $groups = $this->entityTypeManager->getStorage('aws_cloud_security_group')
      ->loadByProperties([
        'cloud_context' => $entity->getCloudContext(),
        'vpc_id' => $entity->getVpcId(),
      ]);

    $options = [];
    array_walk($groups, function ($group) use (&$options) {
      $options[$group->getGroupName()] = $group->getGroupName();
    });

    natcasesort($options);

    $output = [];
    foreach ($options as $key => $value) {
      $output[] = [
        'value' => $key,
        'label' => $value,
      ];
    }

    return new JsonResponse($output, 200);
  }

  /**
   * {@inheritdoc}
   */
  public function getScheduleOptions(): JsonResponse {
    $schedule_dict = aws_cloud_get_schedule();
    $output = [];
    foreach ($schedule_dict as $value => $label) {
      $output[] = [
        'value' => $value,
        'label' => $label,
      ];
    }

    return new JsonResponse($output, 200);
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailableElasticIps(string $cloud_context): JsonResponse {
    $elastic_ips = [];
    $results = $this->entityTypeManager
      ->getStorage('aws_cloud_elastic_ip')
      ->getQuery()
      ->accessCheck(TRUE)
      ->condition('cloud_context', $cloud_context)
      ->notExists('association_id')
      ->execute();

    foreach ($results ?: [] as $result) {
      /** @var \Drupal\aws_cloud\Entity\Ec2\ElasticIp $elastic_ip */
      $elastic_ip = ElasticIp::load($result);
      $elastic_ips[$elastic_ip->getAllocationId()] = $elastic_ip->getPublicIp();
    }

    $output = [];
    foreach ($elastic_ips as $value => $label) {
      $output[] = [
        'value' => $value,
        'label' => $label,
      ];
    }

    return new JsonResponse($output, 200);
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailablePrivateIps(string $entity_id): JsonResponse {
    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
    $entity = $this->entityTypeManager
      ->getStorage('aws_cloud_instance')
      ->load($entity_id);
    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The instance has already been deleted.',
      ], 404);
    }

    $private_ips = [];
    $results = $this->entityTypeManager
      ->getStorage('aws_cloud_network_interface')
      ->getQuery()
      ->accessCheck(TRUE)
      ->condition('instance_id', $entity->getInstanceId())
      ->condition('cloud_context', $entity->getCloudContext())
      ->notExists('association_id')
      ->execute();
    foreach ($results ?: [] as $result) {
      $network_interface = NetworkInterface::load($result);
      if (empty($network_interface)) {
        continue;
      }
      $private_ips[$network_interface->getNetworkInterfaceId()] = $network_interface->getPrimaryPrivateIp();
    }

    $output = [];
    foreach ($private_ips as $value => $label) {
      $output[] = [
        'value' => $value,
        'label' => $label,
      ];
    }

    return new JsonResponse($output, 200);
  }

}
