<?php

namespace Drupal\aws_cloud\Form\Ec2;

use Drupal\aws_cloud\Entity\Ec2\SecurityGroup;
use Drupal\aws_cloud\Plugin\Field\FieldType\IpPermission;
use Drupal\cloud\Entity\CloudContentEntityBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Field\FieldItemList;

/**
 * Form controller for the CloudScripting entity edit forms.
 *
 * @ingroup aws_cloud
 */
class SecurityGroupEditForm extends AwsCloudContentForm {

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $cloud_context = '') {
    $this->ec2Service->setCloudContext($cloud_context);

    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroup $entity */
    $form = parent::buildForm($form, $form_state);

    $entity = $this->entity;
    if (!$this->updateEntityBuildForm($entity)) {
      return $this->redirectUrl;
    }

    $weight = -50;

    $form['security_group'] = [
      '#type'   => 'details',
      '#title'  => $this->t('Security group'),
      '#open'   => TRUE,
      '#weight' => $weight++,
    ];

    $form['security_group']['name'] = [
      '#type'          => 'textfield',
      '#title'         => $this->t('Name'),
      '#maxlength'     => 255,
      '#size'          => 60,
      '#default_value' => $entity->label(),
      '#required'      => TRUE,
      '#weight'        => $weight++,
    ];

    $form['security_group']['group_name'] = [
      '#type'          => 'item',
      '#title'         => $this->getItemTitle($this->t('Security group name')),
      '#markup'        => $entity->getGroupName(),
      '#weight'        => $weight++,
    ];

    $form['security_group']['group_id'] = [
      '#type'          => 'item',
      '#title'         => $this->getItemTitle($this->t('ID')),
      '#markup'        => $entity->getGroupId(),
      '#weight'        => $weight++,
    ];

    $form['security_group']['description'] = [
      '#type'          => 'item',
      '#title'         => $this->getItemTitle($this->t('Description')),
      '#markup'        => $entity->getDescription(),
      '#weight'        => $weight++,
    ];

    $form['security_group']['vpc_id'] = [
      '#type'          => 'item',
      '#title'         => $this->getItemTitle($this->t('VPC ID')),
      '#markup'        => $entity->getVpcId(),
      '#weight'        => $weight++,
    ];

    $form['security_group']['created'] = [
      '#type'          => 'item',
      '#title'         => $this->getItemTitle($this->t('Created')),
      '#markup'        => $this->dateFormatter->format($entity->created(), 'short'),
      '#weight'        => $weight++,
    ];

    // Put all rules into HTML5 details.
    $form['rules'] = [
      '#type' => 'details',
      '#title' => $this->t('Rules'),
      '#open' => TRUE,
      '#weight' => $weight++,
    ];

    $form['rules'][] = $form['ip_permission'];

    if ($entity->getEntityTypeId() === 'aws_cloud_security_group') {
      if (!empty($entity->getVpcId())) {
        $form['rules'][] = $form['outbound_permission'];
      }

      unset($form['ip_permission'], $form['outbound_permission']);
    }
    $form['group_name']['#access'] = FALSE;

    $this->addOthersFieldset($form, $weight++, $cloud_context);

    $form['#attached']['library'][] = 'aws_cloud/aws_cloud_security_groups';

    if (isset($form['actions'])) {
      $form['actions']['submit']['#weight'] = $weight++;
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state): void {
    // Call copyFormItemValues() to ensure the form array is intact.
    $this->copyFormItemValues($form);

    $this->trimTextfields($form, $form_state);

    $entity = $this->entity;

    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $existing_group = $this->entityTypeManager
      ->getStorage($entity->getEntityTypeId())
      ->load($entity->id());

    if ($entity->save()) {

      $this->updateNameAndCreatedByTags($entity, $entity->getGroupId());

      // Update the inbound/outbound permissions.
      $this->updateIngressEgressPermissions($entity, $existing_group);

      // Have the system refresh the security group.
      $this->ec2Service->updateSecurityGroups([
        'GroupIds' => [$this->entity->getGroupId()],
      ], FALSE);

      if (empty($this->messenger->messagesByType('error'))) {
        // Check API calls, see if the permissions updates were
        // successful or not.
        $this->validateAuthorize($entity);
        $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
          'cloud_context' => $entity->getCloudContext(),
          $entity->getEntityTypeId() => $entity->id(),
        ]);
      }
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSaveEvent($entity);
    }
    else {
      $this->processOperationErrorStatus($this->entity, 'updated');
    }
  }

  /**
   * Helper method to update both ingress and egress permissions.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\SecurityGroup $entity
   *   Security group entity.
   * @param \Drupal\aws_cloud\Entity\Ec2\SecurityGroup $existing_group
   *   The exisiting security group entity.
   * @param bool $revokeIngress
   *   Flag indicating whether to skip revoking ingress permissions.
   * @param bool $revokeEgress
   *   Flag indicating whether to skip revoking egress permissions.
   */
  protected function updateIngressEgressPermissions(SecurityGroup $entity, SecurityGroup $existing_group, $revokeIngress = TRUE, $revokeEgress = TRUE): void {
    $inbound_result = $this->updatePermissions(
      $entity->getGroupId(),
      $existing_group->getIpPermission(),
      'IpPermissions',
      'getIpPermission',
      'revokeSecurityGroupIngress',
      'authorizeSecurityGroupIngress',
      $revokeIngress
    );

    // Update the outbound permissions.  This only applies to
    // VPC security groups.
    $outbound_result = TRUE;
    if (!empty($entity->getVpcId())) {
      $outbound_result = $this->updatePermissions(
        $entity->getGroupId(),
        $existing_group->getOutboundPermission(),
        'IpPermissions',
        'getOutboundPermission',
        'revokeSecurityGroupEgress',
        'authorizeSecurityGroupEgress',
        $revokeEgress
      );
    }

    if (!$inbound_result && !$outbound_result) {
      $this->messenger->addError($this->t('Inbound and Outbound Rules could not be saved because of the error from AWS.'));
    }
    elseif (!$inbound_result) {
      $this->messenger->addError($this->t('Inbound Rules could not be saved because of the error from AWS.'));
    }
    elseif (!$outbound_result) {
      $this->messenger->addError($this->t('Outbound Rules could not be saved because of the error from AWS.'));
    }
  }

  /**
   * Main method for updating permissions.
   *
   * Supports Ingress and Egress permissions.
   *
   * @param string $group_id
   *   Security group ID.
   * @param \Drupal\Core\Field\FieldItemList $existing_permissions
   *   Existing permissions.
   * @param string $permission_name
   *   The permission field name.
   * @param string $entity_func
   *   The entity function name to retrieve permissions.  Options are
   *   getIpPermission or getOutboundPermission.
   * @param string $revoke_func
   *   The API function name to revoke permissions.  Options are
   *   revokeSecurityGroupIngress or revokeSecurityGroupEgress.
   * @param string $authorize_func
   *   The API function name to authorize permissions.  Options are
   *   authorizeSecurityGroupIngress or authorizeSecurityGroupEgress.
   * @param bool $revoke
   *   Indicates skipping revoke operation.
   *
   * @return bool
   *   Indicates success or failure
   */
  protected function updatePermissions($group_id,
                                       FieldItemList $existing_permissions,
                                       $permission_name,
                                       $entity_func,
                                       $revoke_func,
                                       $authorize_func,
                                       $revoke = TRUE) : bool {
    // Clear out the existing permissions.
    if (!empty($existing_permissions) && $revoke === TRUE) {
      $revoke_permissions = $this->formatIpPermissionsForRevoke($existing_permissions);
      $this->revokeIpPermissions($revoke_func, $revoke_permissions, $group_id);
    }

    // Format the user entered permissions into AWS accepted array.
    $permissions = $this->formatIpPermissions($this->entity->$entity_func());
    if (!empty($permissions[$permission_name])) {
      // Setup permissions array for AuthorizeSecurityGroupIngress.
      $permissions['GroupId'] = $group_id;
      $result = $this->ec2Service->$authorize_func($permissions);

      // $result === NULL means an exception was thrown.  Attempt to add
      // the old set of permissions so the user's configurations are not lost.
      if ($result === NULL || count($this->messenger->messagesByType('error')) >= 1) {
        if (!empty($existing_permissions)) {
          $existing_permissions = $this->formatIpPermissionsForAdd(
            $existing_permissions,
            $this->entity->getGroupId()
          );
          $result = $this->ec2Service->$authorize_func($existing_permissions);
          if ($result === NULL) {
            $this->messenger->addError($this->t('Cannot re-add existing @name rules', [
              '@name' => $permission_name,
            ]));
          }
        }
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Revoke a set of permissions.
   *
   * @param string $method
   *   String value of revokeSecurityGroupEgress or revokeSecurityGroupIngress.
   * @param array $revoke_permissions
   *   Array of permissions to revoke.
   * @param string $group_id
   *   Security group id to revoke.
   */
  private function revokeIpPermissions($method, array $revoke_permissions, $group_id): void {
    if (!empty($revoke_permissions)) {
      // Delete the existing egress permissions.
      $this->ec2Service->$method([
        'GroupId' => $group_id,
        'IpPermissions' => $revoke_permissions,
      ]);
    }
  }

  /**
   * Helper method to format a permissions array for resubmission.
   *
   * This is used in case of any error during the permissions update process.
   *
   * @param \Drupal\Core\Field\FieldItemList $permissions
   *   IP Permissions.
   * @param string $group_id
   *   Group ID.
   *
   * @return array
   *   Formatted permissions array.
   */
  private function formatIpPermissionsForAdd(FieldItemList $permissions, $group_id): array {
    $existing_permissions['GroupId'] = $group_id;
    foreach ($permissions ?: [] as $permission) {
      if (empty($permission->getSource())) {
        continue;
      }
      $existing_permissions['IpPermissions'][] = $this->formatIpPermissionForAuthorize($permission);
    }
    return $existing_permissions;
  }

  /**
   * Format IP Permissions.
   *
   * @param \Drupal\Core\Field\FieldItemList $ip_permissions
   *   The security group.
   *
   * @return array
   *   Formatted IP Permissions.
   */
  protected function formatIpPermissionsForRevoke(FieldItemList $ip_permissions): array {
    $permissions_formatted = [];
    foreach ($ip_permissions as $ip_permission) {
      if (empty($ip_permission->getSource())) {
        continue;
      }
      $permission_formatted = $this->formatIpPermissionForAuthorize($ip_permission);
      $permissions_formatted[] = $permission_formatted;
    }

    return $permissions_formatted;
  }

  /**
   * Take the ip_permissions field type and format it for AWS API call.
   *
   * @param \Drupal\Core\Field\FieldItemList $permissions
   *   Array of IpPermission fields.
   *
   * @return array
   *   Array of formatted permissions for AWS.
   */
  protected function formatIpPermissions(FieldItemList $permissions): array {
    $ip_permissions = [];
    foreach ($permissions as $permission) {
      if (empty($permission->getSource())) {
        continue;
      }
      $ip_permissions['IpPermissions'][] = $this->formatIpPermissionForAuthorize($permission);
    }
    return $ip_permissions;
  }

  /**
   * Format the IpPermission object.
   *
   * Format the IpPermission object for use with
   * the AuthorizeSecurityGroup [Ingress and Egress] Amazon EC2 API call.
   *
   * @param \Drupal\aws_cloud\Plugin\Field\FieldType\IpPermission $ip_permission
   *   The IP permission object.
   *
   * @return array
   *   The permission.
   */
  protected function formatIpPermissionForAuthorize(IpPermission $ip_permission): array {
    $permission = [
      'FromPort' => (int) $ip_permission->getFromPort(),
      'ToPort' => (int) $ip_permission->getToPort(),
      'IpProtocol' => $ip_permission->getIpProtocol(),
    ];
    if ($ip_permission->getSource() === 'ip4') {
      $item = [
        'CidrIp' => $ip_permission->getCidrIp(),
      ];
      if (!empty($ip_permission->getDescription())) {
        $item['Description'] = $ip_permission->getDescription();
      }
      $permission['IpRanges'][] = $item;
    }
    elseif ($ip_permission->getSource() === 'ip6') {
      $item = [
        'CidrIpv6' => $ip_permission->getCidrIpv6(),
      ];
      if (!empty($ip_permission->getDescription())) {
        $item['Description'] = $ip_permission->getDescription();
      }
      $permission['Ipv6Ranges'][] = $item;
    }
    elseif ($ip_permission->getSource() === 'prefix') {
      $item = [
        'PrefixListId' => $ip_permission->getPrefixListId(),
      ];
      if (!empty($ip_permission->getDescription())) {
        $item['Description'] = $ip_permission->getDescription();
      }
      $permission['PrefixListIds'][] = $item;
    }
    else {
      $group = [];
      // Use GroupID if non-default VPC or EC2-Classic.
      // For other permissions, use Group Name.
      if (!empty($ip_permission->getGroupId())) {
        $group['GroupId'] = $ip_permission->getGroupId();
      }
      else {
        $group['GroupName'] = $ip_permission->getGroupName();
      }

      $group['UserId'] = $ip_permission->getUserId();
      $group['PeeringStatus'] = $ip_permission->getPeeringStatus();
      $group['VpcId'] = $ip_permission->getVpcId();
      $group['VpcPeeringConnectionId'] = $ip_permission->getPeeringConnectionId();
      if (!empty($ip_permission->getDescription())) {
        $group['Description'] = $ip_permission->getDescription();
      }

      $permission['UserIdGroupPairs'][] = $group;
    }
    return $permission;
  }

  /**
   * Verify the authorize call was successful.
   *
   * Since EC2 does not return any error codes from any of the authorize
   * API calls, the only way to verify is to count the permissions array
   * from the current entity,  and the entity that is newly updated from
   * the updateSecurityGroups API call.
   *
   * @param \Drupal\cloud\Entity\CloudContentEntityBase $group
   *   The security group.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  protected function validateAuthorize(CloudContentEntityBase $group): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroup $updated_group */
    $updated_group = $this->entityTypeManager
      ->getStorage($group->getEntityTypeId())
      ->load($group->id());

    $entity = $this->entity;

    if ($group->getIpPermission()->count() !== $updated_group->getIpPermission()->count()) {
      $this->messenger->addError(
        $this->t('Error updating inbound permissions for security group @name', [
          '@name' => $entity->label(),
        ])
      );
    }

    if (!empty($group->getVpcId())) {
      if ($group->getOutboundPermission()->count() !== $updated_group->getOutboundPermission()->count()) {
        $this->messenger->addError(
          $this->t('Error updating outbound permissions for security group @name', [
            '@name' => $entity->label(),
          ])
        );
      }
    }

    if (empty($this->messenger->messagesByType('error'))) {
      // No errors, success.
      $this->processOperationStatus($entity, $this->ec2Service->isWorkerResource() ? 'updated remotely' : 'updated');
    }
  }

}
