<?php

namespace Drupal\aws_cloud\Form\Vpc;

use Drupal\aws_cloud\Form\Ec2\AwsCloudContentForm;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Core\Form\FormStateInterface;

/**
 * Form controller for the transit gateway entity create form.
 *
 * @ingroup aws_cloud
 */
class TransitGatewayCreateForm extends AwsCloudContentForm {

  use CloudContentEntityTrait;

  /**
   * Overrides Drupal\Core\Entity\EntityFormController::buildForm().
   *
   * @param array $form
   *   Array of form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param string $cloud_context
   *   A cloud_context string value from URL "path".
   *
   * @return array
   *   Array of form object.
   */
  public function buildForm(array $form, FormStateInterface $form_state, $cloud_context = ''): array {
    $form = parent::buildForm($form, $form_state);
    $this->ec2Service->setCloudContext($cloud_context);

    $weight = -50;

    $form['transit_gateway'] = [
      '#type' => 'details',
      '#title' => $this->t('Transit gateway'),
      '#open' => TRUE,
      '#weight' => $weight++,
    ];

    $form['transit_gateway']['name'] = [
      '#type'          => 'textfield',
      '#title'         => $this->t('Name'),
      '#maxlength'     => 255,
      '#size'          => 60,
      '#required'      => TRUE,
    ];

    $form['transit_gateway']['description'] = [
      '#type'          => 'textfield',
      '#title'         => $this->t('Description'),
      '#maxlength'     => 255,
      '#size'          => 60,
    ];

    $form['transit_gateway']['amazon_side_asn'] = [
      '#type'          => 'textfield',
      '#title'         => $this->t('Amazon side ASN'),
      '#description'   => $this->t('Enter the Autonomous System Number (ASN) for the AWS side of a Border Gateway Protocol (BGP) session. You can use the default ASN, or you can specify a private ASN in the 64512-65534 or 4200000000-4294967294 ranges.'),
      '#maxlength'     => 255,
      '#size'          => 60,
      '#attributes'    => ['placeholder' => '64512-65534 or 4200000000-4294967294'],
    ];

    $form['transit_gateway']['dns_support'] = [
      '#type'          => 'checkbox',
      '#title'         => $this->t('DNS support'),
      '#description'   => $this->t('Enable Domain Name System resolution for VPCs attached to this transit gateway.'),
      '#default_value' => TRUE,
    ];

    $form['transit_gateway']['vpn_ecmp_support'] = [
      '#type'          => 'checkbox',
      '#title'         => $this->t('VPN ECMP support'),
      '#description'   => $this->t('Equal-cost multi-path routing for VPN Connections that are attached to this transit gateway.'),
      '#default_value' => TRUE,
    ];

    $form['transit_gateway']['default_route_table_association'] = [
      '#type'          => 'checkbox',
      '#title'         => $this->t('Default route table association'),
      '#description'   => $this->t("Automatically associate transit gateway attachments with this transit gateway's default route table."),
      '#default_value' => TRUE,
    ];

    $form['transit_gateway']['default_route_table_propagation'] = [
      '#type'          => 'checkbox',
      '#title'         => $this->t('Default route table propagation'),
      '#description'   => $this->t("Automatically propagate transit gateway attachments with this transit gateway's default route table."),
      '#default_value' => TRUE,
    ];

    $form['transit_gateway']['multicast_support'] = [
      '#type'          => 'checkbox',
      '#title'         => $this->t('Multicast support'),
      '#description'   => $this->t('Enable the ability to create multicast domains in this transit gateway.'),
    ];

    $form['transit_gateway']['auto_accept_shared_attachments'] = [
      '#type'          => 'checkbox',
      '#title'         => $this->t('Auto accept shared attachments'),
      '#description'   => $this->t('Automatically accept cross account attachments that are attached to this transit gateway.'),
    ];

    unset($form['tags']);

    $this->addOthersFieldset($form, $weight++, $cloud_context);

    $form['actions'] = $this->actions($form, $form_state, $cloud_context);

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    parent::validateForm($form, $form_state);

    $amazon_side_asn = $form_state->getValue('amazon_side_asn');
    if (empty($amazon_side_asn)) {
      return;
    }

    $message = $this->t('Specify a private ASN in the 64512-65534 or 4200000000-4294967294 ranges.');
    if (!ctype_digit($amazon_side_asn)) {
      $form_state->setErrorByName('amazon_side_asn', $message);
      return;
    }

    $amazon_side_asn = intval($amazon_side_asn);
    if (!($amazon_side_asn >= 64512 && $amazon_side_asn <= 65534) && !($amazon_side_asn >= 4200000000 && $amazon_side_asn <= 4294967294)) {
      $form_state->setErrorByName('amazon_side_asn', $message);
    }
  }

  /**
   * Overrides Drupal\Core\Entity\EntityFormController::save().
   *
   * @param array $form
   *   Array of form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\aws_cloud\Service\Ec2\Ec2ServiceException
   */
  public function save(array $form, FormStateInterface $form_state): void {
    $this->trimTextfields($form, $form_state);

    $cloud_context = $this->routeMatch->getParameter('cloud_context');

    $entity = $this->entity;
    $entity->setCloudContext($cloud_context);

    $params = [];
    if (!empty($entity->getDescription())) {
      $params['Description'] = $entity->getDescription();
    }

    if (!empty($entity->getAmazonSideAsn())) {
      $params['Options']['AmazonSideAsn'] = $entity->getAmazonSideAsn();
    }
    $params['Options']['AutoAcceptSharedAttachments'] = $entity->isAutoAcceptSharedAttachments() ? 'enable' : 'disable';
    $params['Options']['DefaultRouteTableAssociation'] = $entity->isDefaultRouteTableAssociation() ? 'enable' : 'disable';
    $params['Options']['DefaultRouteTablePropagation'] = $entity->isDefaultRouteTablePropagation() ? 'enable' : 'disable';
    $params['Options']['DnsSupport'] = $entity->isDnsSupport() ? 'enable' : 'disable';
    $params['Options']['MulticastSupport'] = $entity->isMulticastSupport() ? 'enable' : 'disable';
    $params['Options']['VpnEcmpSupport'] = $entity->isVpnEcmpSupport() ? 'enable' : 'disable';
    $result = $this->ec2Service->createTransitGateway($params);

    $form_state->setRedirect(
      'view.aws_cloud_transit_gateway.list',
      ['cloud_context' => $entity->getCloudContext()]
    );

    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'created');
      return;
    }

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'created remotely');
      return;
    }

    $entity->setTransitGatewayId($result['TransitGateway']['TransitGatewayId']);
    $entity->save();

    $this->updateNameAndCreatedByTags($entity, $entity->getTransitGatewayId());

    // Update the transit gateway.
    $this->ec2Service->updateTransitGateways([
      'TransitGatewayIds' => [$entity->getTransitGatewayId()],
    ], FALSE);

    $this->processOperationStatus($entity, 'created');
    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);
  }

}
