<?php

namespace Drupal\aws_cloud\Service\AwsCloud;

use Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface;
use Drupal\aws_cloud\Entity\Ec2\ImageInterface;
use Drupal\aws_cloud\Entity\Ec2\Instance;
use Drupal\aws_cloud\Entity\Ec2\InstanceInterface;
use Drupal\aws_cloud\Form\Ec2\InstanceEditForm;
use Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface;
use Drupal\aws_cloud\Traits\AwsCloudEntityCheckTrait;
use Drupal\cloud\Entity\CloudConfig;
use Drupal\cloud\Entity\CloudConfigInterface;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Core\Entity\EntityDeleteFormTrait;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;

/**
 * Interacts with AWS Cloud using AWS Cloud Service API.
 */
class AwsCloudOperationsService implements AwsCloudOperationsServiceInterface {

  use AwsCloudEntityCheckTrait;
  use CloudContentEntityTrait;
  use EntityDeleteFormTrait;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  private $cloudConfigPluginManager;

  /**
   * The Cloud service.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  private $cloudService;

  /**
   * The AWS Cloud EC2 service.
   *
   * @var \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface
   */
  private $ec2Service;

  /**
   * The entity.
   *
   * @var \Drupal\Core\Entity\EntityInterface
   */
  private $entity;

  /**
   * The Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private $entityTypeManager;

  /**
   * AwsCloudOperationsService constructor.
   *
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   The Cloud service.
   * @param \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface $ec2_service
   *   The AWS Cloud EC2 service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The Entity Type Manager.
   */
  public function __construct(
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    CloudServiceInterface $cloud_service,
    Ec2ServiceInterface $ec2_service,
    EntityTypeManagerInterface $entity_type_manager
  ) {
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->cloudService = $cloud_service;
    $this->ec2Service = $ec2_service;
    $this->entityTypeManager = $entity_type_manager;

    $this->messenger = $this->messenger();
  }

  /**
   * {@inheritdoc}
   */
  public function setEc2Service(mixed $ec2_service): void {
    $this->ec2Service = $ec2_service;
  }

  /**
   * Gets the entity of this method.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   The entity.
   */
  public function getEntity(): EntityInterface {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function copyFormItemValues(FieldableEntityInterface $entity, array $form): void {
    /** @var \Drupal\Core\Entity\FieldableEntityInterface */
    $original_entity = $this->entityTypeManager
      ->getStorage($entity->getEntityTypeId())
      ->load($entity->id());

    $item_field_names = [];
    foreach ($form ?: [] as $name => $item) {
      if (!is_array($item)) {
        continue;
      }

      if (isset($item['#type'])
        && $item['#type'] === 'item'
        && (!isset($item['#not_field']) || $item['#not_field'] === FALSE)
      ) {
        $item_field_names[] = $name;
      }

      if (isset($item['#type']) && $item['#type'] === 'details') {
        foreach ($item ?: [] as $sub_item_name => $sub_item) {
          if (is_array($sub_item)
            && isset($sub_item['#type'])
            && $sub_item['#type'] === 'item'
            && (!isset($sub_item['#not_field']) || $sub_item['#not_field'] === FALSE)
          ) {
            $item_field_names[] = $sub_item_name;
          }
        }
      }
    }

    foreach ($item_field_names ?: [] as $item_field_name) {
      // Support multi-valued item fields.
      $values = !empty($original_entity) ? $original_entity->get($item_field_name)->getValue() : [];
      if (!empty($values) && count($values) > 1) {
        $item_field_values = array_map(static function ($value) {
          return $value['value'];
        }, $values);
        $entity->set($item_field_name, $item_field_values);
      }

      $item_field_value = !empty($original_entity) && !empty($original_entity->get($item_field_name))
        ? $original_entity->get($item_field_name)->value
        : [];
      if (!empty($item_field_value)) {
        $entity->set($item_field_name, $item_field_value);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function trimTextfields(FieldableEntityInterface $entity, array $form, FormStateInterface $form_state): void {
    $field_names = [];
    foreach ($form ?: [] as $name => $item) {
      if (!is_array($item)) {
        continue;
      }

      if (isset($item['#type'])
        && $item['#type'] === 'textfield'
      ) {
        $field_names[] = $name;
      }

      if (isset($item['#type']) && $item['#type'] === 'details') {
        foreach ($item ?: [] as $sub_item_name => $sub_item) {
          if (is_array($sub_item)
            && isset($sub_item['#type'])
            && $sub_item['#type'] === 'textfield'
          ) {
            $field_names[] = $sub_item_name;
          }
        }
      }
    }

    foreach ($field_names ?: [] as $field_name) {
      $value = $form_state->getValue($field_name);
      if ($value === NULL) {
        continue;
      }

      $value = trim($value);
      $form_state->setValue($field_name, $value);
      $entity->set($field_name, $value);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function saveAwsCloudContent(FieldableEntityInterface $entity, array $form, FormStateInterface $form_state): void {
    // Copy values of the form elements whose type are item to entity.
    // If not, the properties corresponding to the form elements
    // will be saved as NULL.
    $this->copyFormItemValues($entity, $form);

    $this->trimTextfields($entity, $form, $form_state);

    $this->saveCloudContent($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function setTagsInAws(EntityInterface $entity, $resource_id, array $tag_map, $entity_type = ''): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $entity */
    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $tags = [];
    foreach ($tag_map ?: [] as $key => $value) {
      $tags[] = [
        'Key' => $key,
        'Value' => $value,
      ];
    }

    if (!method_exists($this->ec2Service, 'createTags')) {
      return;
    }

    // Create Tags with different parameters for AWS and OpenStack.
    if (preg_match('[^aws_cloud]', $entity->getEntityTypeId()) === 1) {
      $this->ec2Service->createTags([
        'Resources' => [$resource_id],
        'Tags' => $tags,
      ]);
    }
    else {
      $this->ec2Service->createTags([
        'Resources' => [$resource_id],
        'Tags' => $tags,
        'EntityType' => $entity_type,
      ]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function updateNameAndCreatedByTags(EntityInterface $entity, $aws_cloud_resource_id): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $entity */
    $cloud_context = $entity->getCloudContext();
    if ($this->isCloudConfigRemote($entity)) {
      // Get original cloud context.
      if (($pos = strrpos($cloud_context, '_')) !== FALSE) {
        $cloud_context = substr($cloud_context, 0, $pos);
      }
    }

    $bundle = str_contains($entity->bundle(), 'aws_cloud')
      ? 'aws_cloud' : 'openstack';
    $uid_key_name = $this->cloudService->getTagCreatedByUid($bundle, $cloud_context);
    $this->setTagsInAws($entity, $aws_cloud_resource_id, [
      $entity->getEntityTypeId() . '_' . $uid_key_name => !empty($entity->getOwner())
      ? $entity->getOwner()->id() : 0,
      'Name' => $entity->getName(),
    ],
    $entity->getEntityTypeId());
  }

  /**
   * {@inheritdoc}
   */
  public function associateElasticIpForInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $allocation_id = $form_state->getValue('allocation_id');
      $network_interface_id = $form_state->getValue('network_interface_id');

      $result = $this->ec2Service->associateAddress([
        'AllocationId' => $allocation_id,
        'NetworkInterfaceId' => $network_interface_id,
      ]);

      if ($result !== NULL) {
        $this->ec2Service->updateElasticIps();
        $this->ec2Service->updateInstances();
        $this->ec2Service->updateNetworkInterfaces();

        $elastic_ip = $this->getElasticIp($allocation_id, $entity->getCloudContext());
        if ($elastic_ip !== FALSE) {
          $this->messenger->addStatus($this->t('Elastic IP @ip_address associated with @private_ip for instance: @instance', [
            '@ip_address' => $elastic_ip->getPublicIp(),
            '@instance' => $entity->getName(),
            '@private_ip' => $elastic_ip->getPrivateIpAddress(),
          ]));
          $this->clearCacheValues($entity->getCacheTags());
        }
        $form_state->setRedirect('entity.aws_cloud_instance.canonical', [
          'cloud_context' => $entity->getCloudContext(),
          'aws_cloud_instance' => $entity->id(),
        ]);
        return TRUE;
      }
      else {
        $this->messenger->addError($this->t('Unable to associate Elastic IP.'));
        $form_state->setRedirect('entity.aws_cloud_instance.canonical', [
          'cloud_context' => $entity->getCloudContext(),
          'aws_cloud_instance' => $entity->id(),
        ]);
        return FALSE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'associated');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createImageFromInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $result = $this->ec2Service->createImage([
        'InstanceId' => $entity->getInstanceId(),
        'Name' => $form_state->getValue('image_name'),
        'NoReboot' => $form_state->getValue('no_reboot') === 0 ? FALSE : TRUE,
      ]);

      if (empty($entity) || empty($result['ImageId'])) {
        // Give an error.
        $this->processOperationErrorStatus($entity, 'created');
        return FALSE;
      }

      // Call image update on this particular image.
      $this->ec2Service->updateImages([
        'ImageIds' => [
          $result['ImageId'],
        ],
      ]);

      $this->messenger()->addStatus($this->t('The @type %label (%image_id) has been created.', [
        '@type' => $entity->getEntityType()->getSingularLabel(),
        '%label' => $entity->toLink($entity->label())->toString(),
        '%image_id' => $result['ImageId'],
      ]));
      $this->clearCacheValues($entity->getCacheTags());
      $image = $this->loadNewImage($entity->getCloudContext(), $result['ImageId']);
      if (!empty($image)) {
        $this->dispatchSubmitEvent($image);
      }
      $form_state->setRedirect('view.aws_cloud_image.list', ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }
      $this->ec2Service->setCloudContext($entity->getCloudContext());

      $result = $this->ec2Service->terminateInstances([
        'InstanceIds' => [$entity->getInstanceId()],
      ]);

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'deleted remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      if (!isset($result['TerminatingInstances'][0]['InstanceId'])) {
        $this->processOperationErrorStatus($entity, 'deleted');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      $entity->delete();

      $this->entity = $entity;
      $this->messenger->addStatus($this->getDeletionMessage());
      $this->logDeletionMessage();
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSubmitEvent($entity);
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function editInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $termination_timestamp = !empty($form_state->getValue('termination_timestamp')[0]['value'])
        ? $form_state->getValue('termination_timestamp')[0]['value']->getTimeStamp()
        : NULL;

      if ($termination_timestamp === NULL) {
        // Unset the termination timestamp.
        $entity->set('termination_timestamp', NULL);
      }

      $security_groups = array_values($form_state->getValue('security_groups'));
      if ($security_groups === NULL) {
        $security_groups = [];
      }
      $entity->setSecurityGroups(implode(InstanceEditForm::SECURITY_GROUP_DELIMITER, $security_groups));

      $this->saveAwsCloudContent($entity, $form, $form_state);

      if (!$form_state->getValue('user_data_base64_encoded')) {
        $entity->setUserData(base64_encode($entity->getUserData() ?? ''));
        $entity->save();
      }

      $user_data_default_value = $form['options']['user_data_base64_encoded']['#default_value']
        ? $form['options']['user_data']['#default_value']
        : base64_encode($form['options']['user_data']['#default_value'] ?? '');

      $config = \Drupal::config('aws_cloud.settings');

      $this->updateTagsField(
        $entity,
        $form_state->getValue('tags'),
        $termination_timestamp,
        $config->get('aws_cloud_scheduler') === TRUE
          ? $entity->getSchedule()
          : NULL
      );

      $this->ec2Service->setCloudContext($entity->getCloudContext());

      $this->updateAwsTags($entity);

      // Update the instance type.
      $this->updateInstanceType($entity, NULL, $form['instance']['instance_type']['#default_value']);

      // Update security group.
      $security_groups = array_values($form_state->getValue('security_groups'));
      if ($security_groups === NULL) {
        $security_groups = [];
      }
      $this->updateSecurityGroup(
        $entity,
        InstanceEditForm::SECURITY_GROUP_DELIMITER,
        $form['network']['security_groups']['#default_value'],
        $security_groups
      );

      // Update terminate protection.
      $this->updateTerminateProtection($entity, NULL, $form['options']['termination_protection']['#default_value']);

      // Update IAM role.
      $this->updateIamRole($entity);

      // Update user data.
      $this->updateUserData($entity, NULL, $user_data_default_value);

      // If Elastic IP is specified and instance is stopped, attach it.
      $this->attachElasticIp(
        $entity,
        $form_state->getValue('add_new_elastic_ip'),
        $form_state->getValue('current_allocation_id'),
        $form_state->getValue('current_association_id')
      );

      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSaveEvent($entity);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'edited');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getConsoleOutput(InstanceInterface $entity): string {
    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $result = $this->ec2Service->getConsoleOutput(['InstanceId' => $entity->getInstanceId()]);

    $output = !empty($result) && !empty($result['Output'])
      ? base64_decode($result['Output'])
      : '';

    return !empty($output) ? $output : '';
  }

  /**
   * {@inheritdoc}
   */
  public function launchInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      // Set parameters to entity.
      $params = [
        'image_id'          => $form['image_id']['#value'],
        'max_count'         => $form['max_count']['#value'],
        'min_count'         => $form['min_count']['#value'],
        'key_pair_name'     => $form['key_pair_name']['#value'],
        'is_monitoring'     => $form['is_monitoring']['#value'] ? TRUE : FALSE,
        'availability_zone' => $form['availability_zone']['#value'],
        'security_groups'   => $form['security_groups']['#value'],
        'instance_type'     => $form['instance_type']['#value'],
        'kernel_id'         => $form['kernel_id']['#value'],
        'ramdisk_id'        => $form['ramdisk_id']['#value'],
        'user_data'         => $form['user_data']['#value'],
      ];

      foreach ($params as $key => $value) {
        $entity->set($key, $value);
      }

      // Get parameters for launching an instance.
      $key_name       = preg_replace('/ \([^)]*\)$/', '', $entity->getKeyPairName());
      $security_group = preg_replace('/ \([^)]*\)$/', '', $entity->getSecurityGroups());

      $params = [
        // The following parameters are required.
        'ImageId'        => $entity->getImageId(),
        'MaxCount'       => $entity->getMaxCount(),
        'MinCount'       => $entity->getMinCount(),
        'InstanceType'   => $entity->getInstanceType(),
        'Monitoring'     => ['Enabled' => $entity->isMonitoring() ? TRUE : FALSE],
        'KeyName'        => $key_name,
        'Placement'      => ['AvailabilityZone' => $entity->getAvailabilityZone()],
        'SecurityGroups' => [$security_group],
      ];

      // The following parameters are optional.
      $params['KernelId'] ?: $entity->getKernelId();
      $params['RamdiskId'] ?: $entity->getRamdiskId();
      $params['UserData'] ?: $entity->getUserData();

      // Launch an instance.
      $result = $this->ec2Service->runInstances($params, []);

      if (!isset($result['Instances'][0]['InstanceId'])) {
        $this->processOperationErrorStatus($entity, 'launched');
        return FALSE;
      }

      if (($entity->setInstanceId($result['Instances'][0]['InstanceId']))
        && ($entity->setPublicIp($result['Instances'][0]['PublicIpAddress']))
        && ($entity->setKeyPairName($result['Instances'][0]['KeyName']))
        && ($entity->setInstanceState($result['Instances'][0]['State']['Name']))
        && ($entity->setCreated($result['Instances'][0]['LaunchTime']))
        && ($entity->save())) {

        $instance_ids = array_map(static function ($result_instance) {
          return $result_instance['InstanceId'];
        }, $result['Instances']);

        $this->messenger->addStatus($this->t('The @type "@label (@instance_id)" request has been initiated. This may take some time. Use Refresh to update the status.', [
          '@type'        => $entity->getEntityType()->getSingularLabel(),
          '@label'       => $entity->label(),
          '@instance_id' => implode(', ', $instance_ids),
        ]));

        $form_state->setRedirectUrl(
          $entity
            ->toUrl('collection')
            ->setRouteParameter('cloud_context', $entity->getCloudContext())
        );
      }
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'launched');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function rebootInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $result = $this->ec2Service->rebootInstances([
        'InstanceIds' => [
          $entity->getInstanceId(),
        ],
      ]);

      if (empty($result)) {
        $this->processOperationErrorStatus($entity, 'rebooted');
        $form_state->setRedirect("view.{$this->entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'rebooted remotely');
        $form_state->setRedirect("view.{$this->entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      $this->processOperationStatus($entity, 'rebooted');
      $this->clearCacheValues($entity->getCacheTags());
      $form_state->setRedirect("view.{$this->entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'rebooted');
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function startInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $result = $this->ec2Service->startInstances([
        'InstanceIds' => [
          $entity->getInstanceId(),
        ],
      ]);

      if (empty($result)) {
        $this->processOperationErrorStatus($entity, 'started');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'started remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'started');
      return FALSE;
    }

    try {
      $current_state = $result['StartingInstances'][0]['CurrentState']['Name'];
      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
      $instance = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());
      $instance->setInstanceState($current_state);
      $instance->save();

      $this->processOperationStatus($instance, 'started');
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSubmitEvent($entity);
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'started');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function stopInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $this->ec2Service->setCloudContext($entity->getCloudContext());
      $result = $this->ec2Service->stopInstances([
        'InstanceIds' => [
          $entity->getInstanceId(),
        ],
      ]);

      if (empty($result)) {
        $this->processOperationErrorStatus($entity, 'stopped');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'stopped remotely');
        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'stopped');
      return FALSE;
    }

    try {
      $current_state = $result['StoppingInstances'][0]['CurrentState']['Name'];
      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
      $instance = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());
      $instance->setInstanceState($current_state);
      $instance->save();

      $this->processOperationStatus($instance, 'stopped');
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSubmitEvent($entity);
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'stopped');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return FALSE;
    }
  }

  /**
   * Save a cloud content.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   Interface for entities having fields.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveCloudContent(FieldableEntityInterface $entity): int {
    if (!($entity instanceof CloudConfigInterface)) {
      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $original_entity */
      $original_entity = $entity;
      $this->cloudConfigPluginManager->setCloudContext($original_entity->getCloudContext());
      $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    }
    else {
      $cloud_config = CloudConfig::create([
        'cloud_context' => $entity->getCloudContext(),
      ]);
    }

    if (!$entity->isNew() && !($entity instanceof CloudConfigInterface) && $cloud_config->isRemote()) {
      $status = SAVED_UPDATED;
      $passive_operation = 'updated remotely';
      $this->processOperationStatus($entity, $passive_operation);
      return $status;
    }

    $status = $entity->save();

    // Add an updated status message and the log.
    switch ($status) {
      case SAVED_NEW:
        $passive_operation = 'created';
        break;

      case SAVED_DELETED:
        $passive_operation = 'deleted';
        break;

      default:
      case SAVED_UPDATED:
        $passive_operation = 'updated';
    }

    if (!($entity instanceof CloudConfigInterface) && $cloud_config->isRemote()) {
      $passive_operation .= ' remotely';
    }

    $this->processOperationStatus($entity, $passive_operation);

    // Clear block and menu cache.
    $this->clearCacheValues($entity->getCacheTags());

    return $status;
  }

  /**
   * Helper function to update field tags.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param array $form_tags
   *   Form tags.
   * @param int|null $termination_timestamp
   *   Termination timestamp by UNIX seconds.
   * @param string|null $schedule
   *   Schedule.
   */
  private function updateTagsField(
    InstanceInterface $entity,
    array $form_tags,
    ?int $termination_timestamp,
    ?string $schedule
  ): void {
    usort($form_tags, static function ($a, $b) {
      return (int) $a['_weight'] - (int) $b['_weight'];
    });

    $module_name = $this->getModuleName($entity);
    $uid_key_name = $this->cloudService->getTagCreatedByUid(
      'aws_cloud',
      $entity->getCloudContext(),
      Instance::TAG_LAUNCHED_BY_UID
    );

    $fixed_tags = [];
    $fixed_tags['Name'] = $entity->getName();
    $fixed_tags[$module_name . '_' . $uid_key_name] = is_null($entity->getOwner())
      ? 0
      : $entity->getOwner()->id();

    $fixed_tags[$module_name . '_' . InstanceInterface::TAG_TERMINATION_TIMESTAMP] = is_null($termination_timestamp)
      ? ''
      : $termination_timestamp;

    $config = \Drupal::config('aws_cloud.settings');
    if ($config->get('aws_cloud_scheduler') === TRUE) {
      $fixed_tags['aws_cloud_scheduler_tag'] = $schedule;
    }

    $tags = [];
    foreach ($form_tags ?: [] as $form_tag) {
      $item_key = $form_tag['item_key'];
      if ($item_key === '') {
        continue;
      }

      // Skip special tags.
      if (strpos($item_key, 'aws:') === 0) {
        continue;
      }

      $item_value = $form_tag['item_value'];
      if (isset($fixed_tags[$item_key])) {
        $item_value = $fixed_tags[$item_key];
        unset($fixed_tags[$item_key]);
      }

      $tags[] = ['item_key' => $item_key, 'item_value' => $item_value];
    }

    foreach ($fixed_tags ?: [] as $item_key => $item_value) {
      $tags[] = ['item_key' => $item_key, 'item_value' => $item_value];
    }

    $entity->setTags($tags);
    $entity->save();
  }

  /**
   * Helper function to update AWS tags.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   */
  private function updateAwsTags(InstanceInterface $entity): void {
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $params = [
      'Resources' => [$entity->getInstanceId()],
    ];

    // Delete old tags.
    $this->ec2Service->deleteTags($params);

    foreach ($entity->getTags() ?: [] as $tag) {
      $params['Tags'][] = [
        'Key' => $tag['item_key'],
        'Value' => $tag['item_value'],
      ];
    }

    // Create Tags with different parameters for AWS and OpenStack.
    if (preg_match('[^aws_cloud]', $entity->getEntityTypeId()) === 1) {
      $this->ec2Service->createTags($params);
    }
    else {
      $this->ec2Service->createTags([
        'Resources' => [$entity->getInstanceId()],
        'Tags' => [
          ['Key' => 'Name', 'Value' => $entity->getName()],
        ],
      ]);
    }
  }

  /**
   * Update the instance type.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param string|null $instance_type
   *   Instance type.
   * @param string|null $default_instance_type
   *   Default value of instance type.
   */
  private function updateInstanceType(
    InstanceInterface $entity,
    ?string $instance_type,
    ?string $default_instance_type
  ): void {
    // If the state of the instance is not stopped, do not update it.
    if ($entity->getInstanceState() !== 'stopped') {
      return;
    }

    // When executed from Form, $instance_type is NULL.
    // When executed from REST API, $instance_type is not NULL.
    // If it is run from Form and the instance type of $entity
    // is equal to the default instance type, the instance type
    // is not changed and no update is performed.
    if ($instance_type === NULL
      && $entity->getInstanceType() === $default_instance_type) {
      return;
    }

    $this->ec2Service->modifyInstanceAttribute([
      'InstanceId' => $entity->getInstanceId(),
      'InstanceType' => ['Value' => $instance_type ?? $entity->getInstanceType()],
    ]);
  }

  /**
   * Get security group IDs by Names.
   *
   * @param array $group_names
   *   Array of group names.
   * @param string $vpc_id
   *   VPC id used for the query.
   *
   * @return array
   *   Array of group names.
   */
  private function getSecurityGroupIdsByNames(array $group_names, string $vpc_id): array {
    // Querying for Group Name is done using filters.
    // @see https://docs.aws.amazon.com/aws-sdk-php/v3/api/api-ec2-2016-11-15.html#describesecuritygroups
    $response = $this->ec2Service->describeSecurityGroups([
      'Filters' => [
        [
          'Name' => 'group-name',
          'Values' => $group_names,
        ],
        [
          // Specify VPC ID because the group names can duplicate across
          // different VPC networks. For example the 'default' group name
          // exists across different groups.
          'Name' => 'vpc-id',
          'Values' => [
            $vpc_id,
          ],
        ],
      ],
    ]);

    $group_ids = [];
    foreach ($response['SecurityGroups'] ?: [] as $security_group) {
      $group_ids[] = $security_group['GroupId'];
    }

    return $group_ids;
  }

  /**
   * Update security group.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param string $security_group_delimiter
   *   Delimiter string for security groups.
   * @param array $default_security_groups
   *   Default value of security groups.
   * @param array $security_groups
   *   Security groups.
   */
  private function updateSecurityGroup(
    InstanceInterface $entity,
    string $security_group_delimiter,
    array $default_security_groups,
    array $security_groups
  ): void {
    if ($entity->getSecurityGroups() !== implode($security_group_delimiter, $default_security_groups)) {

      $this->ec2Service->modifyInstanceAttribute([
        'InstanceId' => $entity->getInstanceId(),
        'Groups' => $this->getSecurityGroupIdsByNames($security_groups, $entity->getVpcId()),
      ]);
    }
  }

  /**
   * Update terminate protection.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param bool|null $termination_protection
   *   Termination protection flag.
   * @param bool|null $default_termination_protection
   *   Default value of termination protection flag.
   */
  private function updateTerminateProtection(
    InstanceInterface $entity,
    ?bool $termination_protection,
    ?bool $default_termination_protection
  ): void {
    if ($termination_protection === NULL
      && $entity->getTerminationProtection() === $default_termination_protection) {
      return;
    }

    $this->ec2Service->modifyInstanceAttribute([
      'InstanceId' => $entity->getInstanceId(),
      'DisableApiTermination' => [
        'Value' => $termination_protection ?? $entity->getTerminationProtection(),
      ],
    ]);
  }

  /**
   * Update IAM role.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   */
  private function updateIamRole(InstanceInterface $entity): void {
    if ($entity->getEntityTypeId() === 'aws_cloud_instance') {
      // Update IAM role.
      $associations_result = $this->ec2Service->describeIamInstanceProfileAssociations([
        'Filters' => [
          [
            'Name' => 'instance-id',
            'Values' => [$entity->getInstanceId()],
          ],
        ],
      ]);

      $associations = $associations_result['IamInstanceProfileAssociations'];
      if (empty($associations)) {
        if ($entity->getIamRole() !== NULL) {
          // Associate.
          $this->ec2Service->associateIamInstanceProfile([
            'InstanceId' => $entity->getInstanceId(),
            'IamInstanceProfile' => [
              'Arn' => $entity->getIamRole(),
            ],
          ]);
        }
      }
      else {
        if ($entity->getIamRole() === NULL) {
          // Disassociate.
          $this->ec2Service->disassociateIamInstanceProfile([
            'AssociationId' => $associations[0]['AssociationId'],
          ]);
        }
        elseif ($entity->getIamRole() !== $associations[0]['IamInstanceProfile']['Arn']) {
          // Disassociate.
          $this->ec2Service->disassociateIamInstanceProfile([
            'AssociationId' => $associations[0]['AssociationId'],
          ]);

          // Associate.
          $this->ec2Service->associateIamInstanceProfile([
            'InstanceId' => $entity->getInstanceId(),
            'IamInstanceProfile' => [
              'Arn' => $entity->getIamRole(),
            ],
          ]);
        }
      }
    }
  }

  /**
   * Update user data.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param string|null $user_data
   *   User data.
   * @param string|null $default_user_data
   *   Default value of user data.
   */
  private function updateUserData(InstanceInterface $entity, ?string $user_data, ?string $default_user_data): void {
    if ($entity->getInstanceState() !== 'stopped') {
      return;
    }

    if ($user_data === NULL
      && $entity->getUserData() === $default_user_data) {
      return;
    }

    $user_data_impl = $user_data ?? base64_decode($entity->getUserData() ?? '');
    $this->ec2Service->modifyInstanceAttribute([
      'InstanceId' => $entity->getInstanceId(),
      'UserData' => ['Value' => $user_data_impl],
    ]);

    $url = Url::fromRoute(
      'entity.aws_cloud_instance.start_form', [
        'cloud_context' => $entity->getCloudContext(),
        'aws_cloud_instance' => $entity->id(),
      ]
    )->toString();

    $this->messenger->addStatus($this->t("The user data of @type %label has been updated. <strong><em><a href='@start_instance_link'>Start</a></em></strong> the @type to reflect the user data.", [
      '@type' => $entity->getEntityType()->getSingularLabel(),
      '%label' => $entity->toLink($entity->label())->toString(),
      '@start_instance_link' => $url,
    ]));
    $this->logOperationMessage($entity, 'updated');
  }

  /**
   * If Elastic IP is specified and instance is stopped, attach it.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   * @param string|null $new_elastic_ip
   *   New Elastic IP.
   * @param string|null $current_allocation_id
   *   Current Allocation ID.
   * @param string|null $current_association_id
   *   Current Association ID.
   */
  private function attachElasticIp(
    InstanceInterface $entity,
    ?string $new_elastic_ip,
    ?string $current_allocation_id,
    ?string $current_association_id
  ): void {
    if ($entity->getInstanceState() === 'stopped') {
      if (isset($new_elastic_ip) && (int) $new_elastic_ip !== -1) {
        $update_entities = FALSE;

        if (isset($current_allocation_id) && isset($current_association_id)) {
          // Reassign the Elastic IP address to the instance.
          // (the Elastic IP address = ['AssociationId', 'InstanceId'])
          if ($current_allocation_id !== $new_elastic_ip) {
            $this->ec2Service->disassociateAddress([
              'AssociationId' => $current_association_id,
              'InstanceId' => $entity->getInstanceId(),
            ]);

            $this->ec2Service->associateAddress([
              'AllocationId' => $new_elastic_ip,
              'InstanceId' => $entity->getInstanceId(),
            ]);
            $update_entities = TRUE;
          }
        }
        else {
          // Instance does not have Elastic IP. Assign a new one.
          $this->ec2Service->associateAddress([
            'AllocationId' => $new_elastic_ip,
            'InstanceId' => $entity->getInstanceId(),
          ]);
          $update_entities = TRUE;
        }
        if ($update_entities === TRUE) {
          // Update the following entities if the Elastic IPs have changed.
          $this->ec2Service->updateElasticIps();
          $this->ec2Service->updateInstances();
          $this->ec2Service->updateNetworkInterfaces();
        }
      }
    }
  }

  /**
   * Helper function that loads an aws_cloud_elastic_ip entity.
   *
   * @param string $allocation_id
   *   The allocation ID to look up.
   * @param string $cloud_context
   *   The cloud context.
   *
   * @return \Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface|false
   *   The loaded aws_cloud_elastic_ip entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getElasticIp(string $allocation_id, string $cloud_context): ElasticIpInterface|FALSE {
    $elastic_ip = FALSE;
    $results = $this->entityTypeManager
      ->getStorage('aws_cloud_elastic_ip')
      ->loadByProperties([
        'allocation_id' => $allocation_id,
        'cloud_context' => $cloud_context,
      ]);
    if (count($results) === 1) {
      $elastic_ip = array_shift($results);
    }
    return $elastic_ip;
  }

  /**
   * Helper method to load image entity.
   *
   * @param string $cloud_context
   *   Cloud context.
   * @param string $image_id
   *   Image id to load.
   *
   * @return null|\Drupal\aws_cloud\entity\ec2\ImageInterface
   *   NULL if not found else Image entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function loadNewImage(string $cloud_context, string $image_id): ?ImageInterface {
    $entities = $this->entityTypeManager
      ->getStorage('aws_cloud_image')
      ->loadByProperties([
        'cloud_context' => $cloud_context,
        'image_id' => $image_id,
      ]);

    return !empty($entities)
      ? array_shift($entities)
      : NULL;
  }

  /**
   * Check if current cloud config is remote or not.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Interface for entities having fields.
   *
   * @return bool
   *   True if the current cloud config is remote.
   */
  private function isCloudConfigRemote(EntityInterface $entity): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $entity */
    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    return $cloud_config->isRemote();
  }

}
