<?php

namespace Drupal\aws_cloud\Service\AwsCloud;

use Drupal\aws_cloud\Entity\Ec2\InstanceInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Provides AwsCloudOperationsService interface.
 */
interface AwsCloudOperationsServiceInterface {

  /**
   * Set the EC2 Service.
   *
   * @param \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface|\Drupal\openstack\Service\OpenStackServiceInterface $ec2_service
   *   The EC2 Service.
   */
  public function setEc2Service(mixed $ec2_service): void;

  /**
   * Copy values from #type=item elements to its original element type.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   Interface for entities having fields.
   * @param array $form
   *   An associative array containing the structure of the form.
   */
  public function copyFormItemValues(FieldableEntityInterface $entity, array $form): void;

  /**
   * Trim white spaces in the values of textfields.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   Interface for entities having fields.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   */
  public function trimTextfields(FieldableEntityInterface $entity, array $form, FormStateInterface $form_state): void;

  /**
   * Save an AWS Cloud content.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   Interface for entities having fields.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function saveAwsCloudContent(FieldableEntityInterface $entity, array $form, FormStateInterface $form_state): void;

  /**
   * Set the tags for the entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Interface for entities having fields.
   * @param string $resource_id
   *   The resource ID.  For example, instance_id, volume_id.
   * @param array $tag_map
   *   The map of tags.
   * @param string $entity_type
   *   The type of entity.
   */
  public function setTagsInAws(EntityInterface $entity, string $resource_id, array $tag_map, string $entity_type = ''): void;

  /**
   * Update name and created_by tags.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Interface for entities having fields.
   * @param string $aws_cloud_resource_id
   *   The AWS Cloud resource ID.
   */
  public function updateNameAndCreatedByTags(EntityInterface $entity, string $aws_cloud_resource_id): void;

  /**
   * Associate an Elastic IP for AWS Cloud instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The instance entity.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function associateElasticIpForInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool;

  /**
   * Create image from AWS Cloud instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The instance entity.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function createImageFromInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool;

  /**
   * Delete an Instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The instance entity.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function deleteInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool;

  /**
   * Edit an Instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The instance entity.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function editInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool;

  /**
   * Get console output from AWS Cloud instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The AWS Cloud instance entity.
   *
   * @return string
   *   Console output.
   */
  public function getConsoleOutput(InstanceInterface $entity): string;

  /**
   * Launch an Instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The instance entity.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function launchInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool;

  /**
   * Reboot an Instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The instance entity.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function rebootInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool;

  /**
   * Start an Instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The instance entity.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function startInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool;

  /**
   * Stop an Instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity
   *   The instance entity.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function stopInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool;

}
