<?php

namespace Drupal\openstack\Service;

use Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface;
use Drupal\aws_cloud\Entity\Ec2\ImageInterface;
use Drupal\aws_cloud\Entity\Ec2\InstanceInterface;
use Drupal\aws_cloud\Entity\Ec2\KeyPairInterface;
use Drupal\aws_cloud\Entity\Ec2\NetworkInterfaceInterface;
use Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface;
use Drupal\aws_cloud\Plugin\Field\FieldType\IpPermission;
use Drupal\aws_cloud\Service\AwsCloud\AwsCloudOperationsServiceInterface;
use Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface;
use Drupal\aws_cloud\Traits\AwsCloudEntityCheckTrait;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\cloud\Entity\CloudContentEntityBase;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityDeleteFormTrait;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\RevisionLogInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\openstack\Entity\OpenStackInstance;
use Drupal\openstack\Service\Rest\OpenStackService as OpenStackRestService;
use Drupal\openstack\Service\Ec2\OpenStackService as OpenStackEc2Service;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Interacts with OpenStack using OpenStackService API.
 */
class OpenStackOperationsService implements OpenStackOperationsServiceInterface {

  use AwsCloudEntityCheckTrait;
  use CloudContentEntityTrait;
  use EntityDeleteFormTrait;

  public const SECURITY_GROUP_DELIMITER = ', ';

  /**
   * The AWS Cloud Operations service.
   *
   * @var \Drupal\aws_cloud\Service\AwsCloud\AwsCloudOperationsServiceInterface
   */
  private $awsCloudOperationsService;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  private $cloudConfigPluginManager;

  /**
   * The Cloud service.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  private $cloudService;

  /**
   * The AWS Cloud EC2 service.
   *
   * @var \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface
   */
  private $ec2Service;

  /**
   * The entity being used by this form.
   *
   * @var \Drupal\Core\Entity\ContentEntityInterface|\Drupal\Core\Entity\RevisionLogInterface
   */
  private $entity;

  /**
   * The Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private $entityTypeManager;

  /**
   * The OpenStack Service Factory.
   *
   * @var \Drupal\openstack\Service\OpenStackServiceFactoryInterface
   */
  private $openStackServiceFactory;

  /**
   * Request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * OpenStackOperationsService constructor.
   *
   * @param \Drupal\aws_cloud\Service\AwsCloud\AwsCloudOperationsServiceInterface $aws_cloud_operations_service
   *   The AWS Cloud Operations service.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   The Cloud service.
   * @param \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface $ec2_service
   *   The AWS Cloud EC2 service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The Entity Type Manager.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   The messenger service.
   * @param \Drupal\openstack\Service\OpenStackServiceFactoryInterface $openstack_service_factory
   *   Object for interfacing with OpenStack Service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   Request stack object.
   */
  public function __construct(
    AwsCloudOperationsServiceInterface $aws_cloud_operations_service,
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    CloudServiceInterface $cloud_service,
    Ec2ServiceInterface $ec2_service,
    EntityTypeManagerInterface $entity_type_manager,
    Messenger $messenger,
    OpenStackServiceFactoryInterface $openstack_service_factory,
    RequestStack $request_stack
  ) {
    $this->awsCloudOperationsService = $aws_cloud_operations_service;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->cloudService = $cloud_service;
    $this->ec2Service = $ec2_service;
    $this->entityTypeManager = $entity_type_manager;
    $this->messenger = $messenger;
    $this->openStackServiceFactory = $openstack_service_factory;
    $this->requestStack = $request_stack;
  }

  /**
   * Gets the form entity.
   */
  public function getEntity(): ContentEntityInterface|RevisionLogInterface {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackInstance(InstanceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      // Switch OpenStack service based on $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      if ($this->ec2Service instanceof OpenStackEc2Service) {
        return $this->awsCloudOperationsService->deleteInstance($this->entity, $form, $form_state);
      }

      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }
      $this->ec2Service->setCloudContext($entity->getCloudContext());

      $result = $this->ec2Service->terminateInstances([
        'InstanceId' => $entity->getInstanceId(),
      ]);

      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

      if (empty($result)) {
        $this->processOperationErrorStatus($entity, 'deleted');
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
          '@type'  => $entity->getEntityType()->getSingularLabel(),
          '@label' => $entity->getName(),
        ]));
        return TRUE;
      }

      $entity->delete();
      $this->messenger->addStatus($this->getDeletionMessage());
      $this->logDeletionMessage();
      $this->clearCacheValues($entity->getCacheTags());

      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function startOpenStackInstance(InstanceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;

      // Switch OpenStack EC2 or Rest service based on
      // $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      $params = [
        'InstanceId' => $entity->getInstanceId(),
      ];

      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

      $result = $this->ec2Service->startInstances($params);
      if (empty($result)) {
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'started remotely');
        return TRUE;
      }

      $current_state = "running";
      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
      $instance = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());
      $instance->setInstanceState($current_state);
      $instance->save();

      $this->processOperationStatus($instance, 'started');
      $this->clearCacheValues($entity->getCacheTags());
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function stopOpenStackInstance(InstanceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;

      // Switch OpenStack EC2 or Rest service based on
      // $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $params = [
        'InstanceId' => $entity->getInstanceId(),
      ];

      $result = $this->ec2Service->stopInstances($params);

      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

      if (empty($result)) {
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'stopped remotely');
        return TRUE;
      }

      $current_state = "stopped";

      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
      $instance = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());
      $instance->setInstanceState($current_state);
      $instance->save();

      $this->processOperationStatus($instance, 'stopped');

      $this->clearCacheValues($entity->getCacheTags());
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Create an Image.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function createImage(array $form, FormStateInterface $form_state): bool {
    $this->awsCloudOperationsService->trimTextfields($this->entity, $form, $form_state);

    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $entity */
    $entity = $this->entity;

    $result = $this->ec2Service->createImage([
      'InstanceId'  => $entity->getInstanceId(),
      'Name'        => $entity->getName(),
      'Description' => $entity->getDescription(),
      'Visibility'  => $entity->getVisibility(),
    ]);

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'created remotely');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }

    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $account_id = $this->cloudConfigPluginManager->loadConfigEntity()->get('field_account_id')->value;

    if (!empty($entity) && !empty($result['ImageId'])
      && ($entity->setName($form_state->getValue('name')))
      && ($entity->set('ami_name', $form_state->getValue('name')))
      && ($entity->setImageId($result['ImageId']))
      && ($entity->set('account_id', $account_id))
      && ($entity->save())) {

      $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getImageId());

      $this->processOperationStatus($entity, 'created');

      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      $this->dispatchSaveEvent($entity);

      return TRUE;
    }
    else {
      $this->processOperationErrorStatus($entity, 'created');

      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackImage(ImageInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;

      // Switch OpenStack service based on $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      $this->createImage($form, $form_state);

      if (!empty($result['SendToWorker'])) {
        return TRUE;
      }

      if ($this->ec2Service instanceof OpenStackRestService) {
        $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getImageId());
      }

      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Save an Image.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveImage(array $form, FormStateInterface $form_state): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface */
    $entity = $this->entity;

    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface */
    $original_entity = $this->entityTypeManager
      ->getStorage($entity->getEntityTypeId())
      ->load($entity->id());

    $this->awsCloudOperationsService->saveAwsCloudContent($entity, $form, $form_state);

    $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getImageId());

    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();

    // Call 'modifyImageAttribute'
    // only when OwnerId is same as Account ID of Cloud Config.
    // It's to avoid AuthFailure error from AWS API.
    if ($cloud_config->get('field_account_id')->value !== $entity->getAccountId()) {
      return;
    }

    $form_values = $form_state->getValues();
    $visibility = $form_values['visibility'] ?? 0;

    $permission = [
      [
        'Group' => 'all',
        'UserId' => $entity->getAccountId(),
      ],
    ];

    // 1 => Public, 0 => Private.
    $launch_permission =
      $visibility
        ? ['Add' => $permission]
        : ['Remove' => $permission];

    // Update image.
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    if (!method_exists($this->ec2Service, 'modifyImageAttribute')) {
      return;
    }

    // Update image description.
    $this->ec2Service->modifyImageAttribute([
      'ImageId' => $entity->getImageId(),
      'Description' => ['Value' => $entity->getDescription()],
    ]);

    // These fields (Description , LaunchPermission , or ProductCode) cannot
    // be specified at the same time.
    // See also:
    // https://docs.aws.amazon.com/cli/latest/reference/ec2/modify-image-attribute.html
    // Update image launch permission.
    $this->ec2Service->modifyImageAttribute([
      'ImageId' => $entity->getImageId(),
      'LaunchPermission' => $launch_permission,
    ]);

    $new_account_ids = [];
    foreach ($entity->getLaunchPermissionAccountIds() as $account_id) {
      if (empty($account_id->value)) {
        continue;
      }
      $new_account_ids[] = $account_id->value;
    }

    $old_account_ids = [];
    foreach ($original_entity->getLaunchPermissionAccountIds() as $account_id) {
      if (empty($account_id->value)) {
        continue;
      }
      $old_account_ids[] = $account_id->value;
    }

    $account_ids_to_remove = array_diff($old_account_ids, $new_account_ids);
    $account_ids_to_add = array_diff($new_account_ids, $old_account_ids);

    if (!empty($account_ids_to_remove)) {
      $this->ec2Service->modifyImageAttribute([
        'ImageId' => $entity->getImageId(),
        'LaunchPermission' => [
          'Remove' => array_map(function ($item) {
            return ['UserId' => $item];
          }, $account_ids_to_remove),
        ],
      ]);
    }

    if (!empty($account_ids_to_add)) {
      $this->ec2Service->modifyImageAttribute([
        'ImageId' => $entity->getImageId(),
        'LaunchPermission' => [
          'Add' => array_map(function ($item) {
            return ['UserId' => $item];
          }, $account_ids_to_add),
        ],
      ]);
    }

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackImage(ImageInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      $this->saveImage($form, $form_state);

      $visibility = !empty($entity->getVisibility()) ? 'public' : 'private';

      // Update Image of OpenStack REST API.
      if ($this->ec2Service instanceof OpenStackRestService) {
        // Make an API call to update image.
        $result = $this->ec2Service->updateImage([
          'ImageId' => $entity->getImageId(),
          'Name' => $entity->getName(),
          'Visibility' => $visibility,
        ]);

        $old_tags = $result['Tags'] ?? [];

        foreach ($old_tags ?: [] as $tag) {
          // Delete old tags.
          $this->ec2Service->deleteTags(
            [
              'ImageId'    => $entity->getImageId(),
              'Tag'        => $tag,
              'EntityType' => $entity->getEntityTypeId(),
            ]);
        }

        $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getImageId());
      }

      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Delete an Image.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function deleteImage(FormStateInterface $form_state): bool {
    /**
     * The entity being used by this form.
     *
     * @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface|\Drupal\aws_cloud\Entity\Ec2\Image
     */
    $entity = $this->entity;
    if ($entity->getEntityTypeId() === 'aws_cloud_image'
        && !$this->updateEntitySubmitForm($form_state, $entity)) {
      return FALSE;
    }

    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $account_id = $this->cloudConfigPluginManager->loadConfigEntity()->get('field_account_id')->value;

    // If the image is not owned by the aws user,
    // the calling for deregisterImage will be skipped for AWS Cloud,
    // and it will be called for OpenStack.
    if (($entity->getEntityTypeId() === 'aws_cloud_image'
    && ($entity->getAccountId() !== $account_id))
    || ($result = $this->ec2Service->deregisterImage([
      'ImageId' => $entity->getImageId(),
    ])) !== NULL) {

      if (!empty($result['SendToWorker'])) {
        $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
          '@type'  => $entity->getEntityType()->getSingularLabel(),
          '@label' => $entity->getName(),
        ]));

        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      $entity->delete();

      // Do not change the following message since we cannot use
      // $entity->label(), which represents 'ami_name'.
      $this->messenger->addStatus($this->t('The @type @label has been deleted.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));
      $this->logDeletionMessage();
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSubmitEvent($entity);
    }
    else {

      // Do not change the following message since we cannot use
      // $entity->label(), which represents 'ami_name'.
      $this->messenger->addError($this->t('The @type @label could not be deleted.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));

      $this->logger($entity->getEntityType()->getProvider())->error($this->t('@type: @label could not be deleted.', [
        '@type' => $entity->getEntityType()->getLabel(),
        '@label' => $entity->label(),
      ]), [
        'link' => $entity->toLink($this->t('View'))->toString(),
      ]);
    }

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackImage(ImageInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->deleteImage($form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackKeyPair(KeyPairInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->createKeyPair($form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Create a KeyPair.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function createKeyPair(array &$form, FormStateInterface $form_state): bool {
    $this->awsCloudOperationsService->trimTextfields($this->entity, $form, $form_state);

    /** @var \Drupal\aws_cloud\Entity\Ec2\KeyPair $entity */
    $entity = $this->entity;
    $uid_key_name = $this->cloudService->getTagCreatedByUid(
      'aws_cloud',
      $entity->getCloudContext());

    $result = $this->ec2Service->createKeyPair([
      'KeyName' => $entity->getKeyPairName(),
      'TagSpecifications' => [
        [
          'ResourceType' => 'key-pair',
          'Tags' => [
            [
              'Key' => $entity->getEntityTypeId() . '_' . $uid_key_name,
              'Value' => !empty($entity->getOwner()) ? $entity->getOwner()->id() : 0,
            ],
          ],
        ],
      ],
    ]);

    if (empty($result)) {
      // Use the custom message since key pair uses 'key_pair_name' for its own
      // label.  So do not change the following code.
      $this->messenger->addStatus($this->t('The @type @label could not be created.', [
        '@type' => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getKeyPairName(),
      ]));
      $this->logOperationErrorMessage($entity, 'created');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'created remotely');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }

    $entity->setKeyFingerprint($result['KeyFingerprint']);
    $entity->setKeyPairId($result['KeyPairId']);
    $entity->setKeyMaterial($result['KeyMaterial']);
    $entity->save();
    $this->processOperationStatus($entity, 'created');
    $this->logOperationMessage($entity, 'created');
    $this->clearCacheValues($entity->getCacheTags());

    $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
      'cloud_context' => $entity->getCloudContext(),
      $entity->getEntityTypeId() => $entity->id(),
    ]);
    $this->dispatchSaveEvent($entity);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackKeyPair(KeyPairInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      $this->saveKeyPair($form, $form_state);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Save a Key Pair.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveKeyPair(array $form, FormStateInterface $form_state): void {
    // Call copyFormItemValues() to ensure the form array is intact.
    $this->awsCloudOperationsService->copyFormItemValues($this->entity, $form);

    $this->awsCloudOperationsService->trimTextfields($this->entity, $form, $form_state);

    /**
     * The entity being used by this form.
     *
     * @var \Drupal\aws_cloud\Entity\Ec2\KeyPairInterface|\Drupal\aws_cloud\Entity\Ec2\KeyPair
     */
    $entity = $this->entity;

    $bundle = str_contains($entity->bundle(), 'aws_cloud')
      ? 'aws_cloud' : 'openstack';

    $uid_key_name = $this->cloudService->getTagCreatedByUid(
      $bundle,
      $entity->getCloudContext());

    $this->ec2Service->setCloudContext($entity->getCloudContext());

    if ($entity->save()) {
      $this->awsCloudOperationsService->setTagsInAws($entity, $entity->getKeyPairName(), [
        $entity->getEntityTypeId() . '_' . $uid_key_name => !empty($entity->getOwner())
        ? $entity->getOwner()->id() : 0,
      ], $entity->getEntityTypeId());
    }

    $this->processOperationStatus($entity, 'updated');
    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackKeyPair(KeyPairInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->deleteKeyPair($form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Delete a KeyPair.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function deleteKeyPair(FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\KeyPairInterface|\Drupal\aws_cloud\Entity\Ec2\KeyPair */
    $entity = $this->entity;
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $result = $this->ec2Service->deleteKeyPair([
      'KeyName' => $entity->getKeyPairName(),
    ]);

    // If $result is NULL, some error should have occurred
    // when calling the AWS API.
    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));
      return TRUE;
    }

    $entity->delete();

    $this->messenger->addStatus($this->getDeletionMessage());
    $this->logDeletionMessage();
    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSubmitEvent($entity);
    return TRUE;
  }

  /**
   * Helper function to update field tags.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   */
  private function updateTagsField(FormStateInterface $form_state): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
    $entity = $this->entity;

    $form_tags = $form_state->getValue('tags');
    usort($form_tags, static function ($a, $b) {
      return (int) $a['_weight'] - (int) $b['_weight'];
    });
    $uid_key_name = $this->cloudService->getTagCreatedByUid(
      'openstack',
      $entity->getCloudContext(),
      OpenStackInstance::TAG_LAUNCHED_BY_UID
    );

    $module_name = $this->getModuleName($entity);

    $fixed_tags = [];
    $fixed_tags[$module_name . '_' . $uid_key_name] = $entity->getOwner() === NULL
    ? 0
    : $entity->getOwner()->id();

    $tags = [];
    foreach ($form_tags ?: [] as $form_tag) {
      $item_key = $form_tag['item_key'];
      if ($item_key === '') {
        continue;
      }

      // Skip special tags.
      if (strpos($item_key, 'openstack:') === 0) {
        continue;
      }

      $item_value = $form_tag['item_value'];

      if ($form_tag['item_key'] !== $module_name . '_' . $uid_key_name) {
        $tags[] = ['item_key' => $item_key, 'item_value' => $item_value];
      }
    }

    if (!empty($fixed_tags)) {
      foreach ($fixed_tags ?: [] as $item_key => $item_value) {
        if ($item_value) {
          $tags[] = ['item_key' => $item_key, 'item_value' => $item_value];
        }
      }
    }

    $entity->setTags($tags);
    $entity->save();
  }

  /**
   * Helper function to update OpenStack tags.
   */
  private function updateOpenStackTags(): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
    $entity = $this->entity;

    $params = [
      'Resources' => [$entity->getInstanceId()],
    ];

    // Delete old tags.
    $this->ec2Service->deleteTags(
      [
        'InstanceId' => $entity->getInstanceId(),
        'EntityType' => $entity->getEntityTypeId(),
      ]);

    foreach ($entity->getTags() ?: [] as $tag) {
      $params['Tags'][] = [
        'Key' => $tag['item_key'],
        'Value' => $tag['item_value'],
      ];
    }

    // Create Tags with different parameters for OpenStack.
    $this->ec2Service->createTags([
      'Resources' => [$entity->getInstanceId()],
      'Tags' => $params['Tags'],
      'EntityType' => $entity->getEntityTypeId(),
    ]);

    $this->ec2Service->updateInstances();

  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->awsCloudOperationsService->copyFormItemValues($this->entity, $form);
      // Switch OpenStack EC2 or Rest service
      // based on $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      if (!($this->ec2Service instanceof OpenStackRestService)) {
        return $this->awsCloudOperationsService->editInstance($this->entity, $form, $form_state);
      }

      $termination_timestamp = $form_state->getValue('termination_timestamp')[0]['value'] ?? '';

      $this->updateTagsField($form_state);

      $this->updateOpenStackTags();

      if ($termination_timestamp === NULL) {
        // Unset the termination timestamp.
        $entity->set('termination_timestamp', NULL);
      }

      // Update Instance of OpenStack REST API.
      $result = $this->ec2Service->modifyLaunchTemplate([
        'InstanceId' => $entity->getInstanceId(),
        'Name' => $entity->getName(),
      ]);

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'updated remotely');
        return TRUE;
      }

      $entity->save();

      $this->processOperationStatus($entity, 'updated');

      $this->clearCacheValues($entity->getCacheTags());
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackNetworkInterface(NetworkInterfaceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->createNetworkInterface($form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Create a NetworkInterface.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function createNetworkInterface(array &$form, FormStateInterface $form_state): bool {
    $this->awsCloudOperationsService->trimTextfields($this->entity, $form, $form_state);

    /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterfaceInterface|\Drupal\aws_cloud\Entity\Ec2\NetworkInterface */
    $entity = $this->entity;
    $security_group_ids = !empty($form_state->getValue('security_groups'))
      ? array_values($form_state->getValue('security_groups')) : [];

    $security_groups = !empty($security_group_ids)
      ? implode(',', $security_group_ids) : '';
    $entity->setSecurityGroups($security_groups);

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $result = $this->ec2Service->createNetworkInterface([
      'Name' => $entity->getName(),
      'Description' => $entity->getDescription(),
      'SubnetId' => $entity->getSubnetId(),
      'Groups' => $security_group_ids,
    ]);

    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'created remotely');
      return TRUE;
    }

    $entity->setNetworkInterfaceId($result['NetworkInterface']['NetworkInterfaceId']);
    $entity->save();

    if ($this->ec2Service instanceof OpenStackRestService) {
      $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getNetworkInterfaceId());
    }

    // Update the VPC.
    $this->ec2Service->updateNetworkInterfaces([
      'NetworkInterfaceId' => $entity->getNetworkInterfaceId(),
    ]);

    $this->processOperationStatus($entity, 'created');

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackNetworkInterface(NetworkInterfaceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterface */
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($this->entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->deleteNetworkInterface($form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Delete a NetworkInterface.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function deleteNetworkInterface(FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterface $entity */
    $entity = $this->entity;

    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $result = $this->ec2Service->deleteNetworkInterface([
      'NetworkInterfaceId' => $entity->getNetworkInterfaceId(),
    ]);

    // If $result is NULL, some error should have occurred
    // when calling the AWS API.
    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));
      return TRUE;
    }

    $entity->delete();

    $this->messenger->addStatus($this->getDeletionMessage());
    $this->logDeletionMessage();
    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSubmitEvent($entity);

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackNetworkInterface(NetworkInterfaceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterface */
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($this->entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      $this->saveNetworkInterface($form, $form_state);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Save a Network Interface.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveNetworkInterface(array $form, FormStateInterface $form_state): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterface */
    $entity = $this->entity;

    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $this->awsCloudOperationsService->saveAwsCloudContent($entity, $form, $form_state);

    $params = [
      'NetworkInterfaceId' => $entity->getNetworkInterfaceId(),
      'Description' => ['Value' => $entity->getDescription()],
    ];

    if (method_exists($this->ec2Service, 'modifyNetworkInterfaceAttribute')) {
      $this->ec2Service->modifyNetworkInterfaceAttribute($params);
    }

    $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getNetworkInterfaceId());

    // Update OpenStack REST NetworkInterfaces.
    if ($this->ec2Service instanceof OpenStackRestService) {
      // Delete old tags.
      $this->ec2Service->deleteTags(
        [
          'NetworkInterfaceId' => $entity->getNetworkInterfaceId(),
          'EntityType'         => $entity->getEntityTypeId(),
        ]);

      $this->ec2Service->updateNetworkInterface([
        'NetworkInterfaceId' => $entity->getNetworkInterfaceId(),
        'Name' => $entity->getName(),
        'Description' => $entity->getDescription(),
      ]);
    }

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackFloatingIp(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;

    // Switch OpenStack EC2 or REST service based on $entity->getCloudContext().
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

    if (!($this->ec2Service instanceof OpenStackRestService)) {
      return $this->createFloatingIp($form, $form_state);
    }

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
    $params = [
      'AllocationId' => $form_state->getValue('floating_network_id'),
    ];

    $result = $this->ec2Service->allocateAddress($params);

    if (empty($result)) {
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'created remotely');
      return TRUE;
    }

    $floating_network_id = $form_state->getValue('floating_network_id');
    $entity->setAllocationId($floating_network_id);
    $entity->save();

    $public_ip = $result['PublicIp'] ?? NULL;
    $floating_ip_type = !empty($result['PublicIp']) ? 'Public IP' : '';
    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    if (!empty($public_ip)
      && $entity->setFloatingIpId($result['FloatingIpId'] ?? '')
      && $entity->setPublicIp($public_ip)
      && $entity->setElasticIpType($floating_ip_type)
      && $entity->save()) {

      // Create uid tag.
      $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getFloatingIpId());
      $this->processOperationStatus($entity, 'created');
      $this->clearCacheValues($entity->getCacheTags());
    }
    else {
      $this->processOperationErrorStatus($entity, 'created');
    }
    return TRUE;
  }

  /**
   * Create a floating IP.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function createFloatingIp(array &$form, FormStateInterface $form_state): bool {
    $this->awsCloudOperationsService->trimTextfields($this->entity, $form, $form_state);

    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp */
    $entity = $this->entity;

    $params = [
      'Domain' => $entity->getDomain(),
    ];

    if (!empty($entity->getNetworkBorderGroup())) {
      $params['NetworkBorderGroup'] = $entity->getNetworkBorderGroup();
    }

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $result = $this->ec2Service->allocateAddress($params);
    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'created remotely');
      return TRUE;
    }

    $public_ip = $result['PublicIp'] ?? $result['CarrierIp'];
    $elastic_ip_type = !empty($result['PublicIp']) ? 'Public IP' : 'Carrier IP';
    $entity->setPublicIp($public_ip);
    $entity->setAllocationId($result['AllocationId']);
    $entity->setDomain($result['Domain']);
    $entity->setElasticIpType($elastic_ip_type);
    $entity->save();

    if (!empty($result['NetworkBorderGroup'])) {
      $entity->setNetworkBorderGroup($result['NetworkBorderGroup']);
      $entity->save();
    }

    // Update the entity name.
    $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getAllocationId());

    $this->processOperationStatus($entity, 'created');
    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackFloatingIp(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;

    // Copy values of the form elements whose type are item to entity.
    // If not, the properties corresponding to the form elements
    // will be saved as NULL.
    $this->awsCloudOperationsService->copyFormItemValues($this->entity, $form);

    $this->awsCloudOperationsService->trimTextfields($this->entity, $form, $form_state);

    // Switch OpenStack EC2 or REST service based on $entity->getCloudContext().
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    if (!$this->ec2Service instanceof OpenStackRestService) {
      $this->saveFloatingIp($form, $form_state);
      return TRUE;
    }

    // Delete old tags.
    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $this->ec2Service->deleteTags(
      [
        'FloatingIpId' => $entity->getFloatingIpId(),
        'EntityType'   => $entity->getEntityTypeId(),
      ]);

    $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getFloatingIpId());

    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    if ($cloud_config->isRemote()) {
      $this->processOperationStatus($entity, 'updated remotely');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }

    if ($entity->save()) {
      $this->processOperationStatus($entity, 'updated');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }

    $this->processOperationErrorStatus($entity, 'updated');
    return TRUE;
  }

  /**
   * Save a floating IP.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveFloatingIp(): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ElasticIp */
    $entity = $this->entity;

    $this->entity->save();

    $this->ec2Service->setCloudContext($entity->getCloudContext());

    // Update the name.
    $this->awsCloudOperationsService->updateNameAndCreatedByTags($this->entity, $entity->getAllocationId());

    $this->processOperationStatus($this->entity, 'updated');

    $this->clearCacheValues($this->entity->getCacheTags());
    $this->dispatchSaveEvent($this->entity);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackFloatingIp(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());

    if ($this->ec2Service instanceof OpenStackEc2Service) {
      return $this->deleteFloatingIp($form_state);
    }

    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $params = ['FloatingIpId' => $entity->getFloatingIpId()];

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    if (empty($entity)
      || empty($params)
      || empty($this->ec2Service->releaseAddress($params))) {
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }

    $entity->delete();

    $this->messenger->addStatus($this->getDeletionMessage());
    $this->logDeletionMessage();
    $this->clearCacheValues($entity->getCacheTags());
    return TRUE;
  }

  /**
   * Delete a floating IP.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function deleteFloatingIp(FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ElasticIp $entity */
    $entity = $this->entity;
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $allocation_id = $entity->getAllocationId();
    $public_ip = $entity->getPublicIp();
    $params = [];
    if ($entity->getDomain() === 'standard' && !empty($public_ip)) {
      $params['PublicIp'] = $public_ip;
    }
    elseif ($entity->getDomain() === 'vpc' && !empty($allocation_id)) {
      $params['AllocationId'] = $allocation_id;
      $params['NetworkBorderGroup'] = $entity->getNetworkBorderGroup();
    }

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $result = $this->ec2Service->releaseAddress($params);

    // If $result is NULL, some error should have occurred
    // when calling the AWS API.
    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));
      return TRUE;
    }

    // Update instances after the Elastic IP is deleted.
    if ($entity->getEntityTypeId() === 'aws_cloud_elastic_ip') {
      $this->ec2Service->updateInstances();
    }

    $entity->delete();

    $this->messenger->addStatus($this->getDeletionMessage());
    $this->logDeletionMessage();
    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSubmitEvent($entity);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackSecurityGroup(SecurityGroupInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->createSecurityGroup($form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Create a SecurityGroup.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function createSecurityGroup(array &$form, FormStateInterface $form_state): bool {
    $this->awsCloudOperationsService->trimTextfields($this->entity, $form, $form_state);

    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroup */
    $entity = $this->entity;

    if ($entity->getEntityTypeId() === 'aws_cloud_security_group') {
      $result = $this->ec2Service->createSecurityGroup([
        'GroupName'   => $entity->getGroupName(),
        'VpcId'       => $entity->getVpcId(),
        'Description' => $entity->getDescription(),
      ]);
    }
    else {
      $result = $this->ec2Service->createSecurityGroup([
        'GroupName'   => $entity->getGroupName(),
        'Description' => $entity->getDescription(),
      ]);
    }

    if (!empty($result['SendToWorker'])) {
      $entity->setName($entity->getGroupName());
      $this->processOperationStatus($entity, 'created remotely');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }

    if (isset($result['GroupId'])
    && ($entity->setGroupId($result['GroupId']))
    && ($entity->set('name', $entity->getGroupName()))
    && ($entity->save())) {
      $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getGroupId());

      // Have the system refresh the security group.
      $this->ec2Service->updateSecurityGroups([
        'GroupIds' => [$entity->getGroupId()],
      ], FALSE);

      $this->processOperationStatus($entity, 'created');
      $this->messenger->addStatus('Setup the IP permissions.');

      $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
        'cloud_context' => $entity->getCloudContext(),
        $entity->getEntityTypeId() => $entity->id(),
      ]);
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSaveEvent($entity);
    }
    else {
      // Use the custom message since security group uses 'group_name' for its
      // own label.  So do not change the following code.
      $message = $this->t('The @type @label could not be created.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getGroupName(),
      ]);
      $this->messenger->addError($message);
      $this->logOperationErrorMessage($entity, 'created');
      return FALSE;
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackSecurityGroup(SecurityGroupInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      return $this->deleteSecurityGroup($form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Delete a SecurityGroup.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function deleteSecurityGroup(FormStateInterface $form_state): bool {

    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroup */
    $entity = $this->entity;
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $result = $this->ec2Service->deleteSecurityGroup([
      'GroupId'   => $entity->getGroupId(),
    ]);

    // If $result is NULL, some error should have occurred
    // when calling the AWS API.
    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));
      return TRUE;
    }

    $entity->delete();

    $this->messenger->addStatus($this->getDeletionMessage());
    $this->logDeletionMessage();
    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSubmitEvent($entity);

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function revokeOpenStackSecurityGroup(
    SecurityGroupInterface $entity,
    array &$form,
    FormStateInterface $form_state
  ): bool {
    $this->entity = $entity;
    // Switch OpenStack EC2 or REST service based on $entity->getCloudContext().
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());

    $referer = $this->requestStack->getCurrentRequest()->headers->get('referer');
    $url = parse_url($referer);
    parse_str($url['query'], $results);
    $type = $results['type'];
    $position = $results['position'];

    // Use parent save function in case of OpenStack EC2.
    if (!($this->ec2Service instanceof OpenStackRestService)) {
      return $this->revokeSecurityGroup($form_state, $type, $position);
    }

    /** @var \Drupal\cloud\Entity\CloudContentEntityBase|\Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $entity */
    $permission = $this->getPermission($type, $position, $entity);
    if ($permission !== FALSE) {

      $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
        'cloud_context' => $entity->getCloudContext(),
        $entity->getEntityTypeId() => $entity->id(),
      ]);

      // Revoke security group rule.
      $perm_array['IpPermissions'][] = [
        'RuleId' => $permission->rule_id,
      ];
      $param_array = [
        'RuleId' => $perm_array['IpPermissions'][0]['RuleId'],
      ];
      $result = $this->ec2Service->revokeSecurityGroupRule($param_array);

      if (!empty($result['SendToWorker'])) {
        $this->messenger->addStatus($this->t('Permission revoked remotely'));
        return TRUE;
      }

      // Have the system refresh the security group.
      $this->ec2Service->updateSecurityGroups([
        'GroupIds' => [$entity->getGroupId()],
      ], FALSE);

      $this->validateRevoke($type, $entity)
        ? $this->messenger->addStatus($this->t('Permission revoked'))
        : $this->messenger->addError($this->t('Permission not revoked'));

      $this->clearCacheValues($entity->getCacheTags());
    }

    return TRUE;
  }

  /**
   * Revoke a Rest SecurityGroup.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param string $type
   *   Outbound_permission or ip_permission.
   * @param string $position
   *   Position in the multivalued list of permissions.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function revokeSecurityGroup(FormStateInterface $form_state, string $type, string $position): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface|\Drupal\cloud\Entity\CloudContentEntityBase */
    $entity = $this->entity;
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $permission = $this->getPermission($type, $position, $entity);
    if (!$permission) {
      return FALSE;
    }

    $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
      'cloud_context' => $entity->getCloudContext(),
      $entity->getEntityTypeId() => $entity->id(),
    ]);

    $perm_array = [
      'GroupId' => $entity->getGroupId(),
    ];

    if ($type === 'outbound_permission') {
      $perm_array['IpPermissions'][] = $this->formatIpPermissionForRevoke($permission);
      $result = $this->ec2Service->revokeSecurityGroupEgress($perm_array);
    }
    else {
      $perm_array['IpPermissions'][] = $this->formatIpPermissionForRevoke($permission);
      $result = $this->ec2Service->revokeSecurityGroupIngress($perm_array);
    }

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('Permission revoked remotely'));
      return TRUE;
    }

    // Have the system refresh the security group.
    $this->ec2Service->updateSecurityGroups([
      'GroupIds' => [$entity->getGroupId()],
    ], FALSE);

    $this->validateRevoke($type, $entity)
      ? $this->messenger->addStatus($this->t('Permission revoked'))
      : $this->messenger->addError($this->t('Permission not revoked'));

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSubmitEvent($entity);

    return TRUE;
  }

  /**
   * Get an individual permission from outbound or inbound permission.
   *
   * @param string $type
   *   Outbound_permission or ip_permission.
   * @param int $position
   *   Position in the multivalued list of permissions.
   * @param \Drupal\cloud\Entity\CloudContentEntityBase $entity
   *   Security group entity.
   *
   * @return bool|object
   *   FALSE or permission object.
   */
  private function getPermission(string $type, int $position, CloudContentEntityBase $entity): ?object {
    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface|\Drupal\cloud\Entity\CloudContentEntityBase $entity */

    $permission = FALSE;
    switch ($type) {
      case 'outbound_permission':
        $permission = $entity->getOutboundPermission()->get($position);
        break;

      case 'ip_permission':
        $permission = $entity->getIpPermission()->get($position);
        break;

      default:
        break;
    }
    return $permission;
  }

  /**
   * Format the IP permission array for use with revoke security group.
   *
   * @param \Drupal\aws_cloud\Plugin\Field\FieldType\IpPermission $ip_permission
   *   IP Permission object.
   *
   * @return array
   *   Permissions array.
   */
  private function formatIpPermissionForRevoke(IpPermission $ip_permission): array {
    $permission = [
      'FromPort' => (int) $ip_permission->getFromPort(),
      'ToPort' => (int) $ip_permission->getToPort(),
      'IpProtocol' => $ip_permission->getIpProtocol(),
    ];
    if ($ip_permission->getSource() === 'ip4') {
      $permission['IpRanges'][]['CidrIp'] = $ip_permission->getCidrIp();
    }
    else {
      if ($ip_permission->getSource() === 'ip6') {
        $permission['Ipv6Ranges'][]['CidrIpv6'] = $ip_permission->getCidrIpv6();
      }
      else {
        // Use GroupID if non-default VPC or EC2-Classic.
        // For other permissions, use Group Name.
        /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface */
        $entity = $this->entity;
        $vpc_id = $entity->getVpcId();
        if ($entity->isDefaultVpc() === TRUE || !isset($vpc_id)) {
          $group['GroupName'] = $ip_permission->getGroupName();
        }
        else {
          $group['GroupId'] = $ip_permission->getGroupId();
        }
        $group['UserId'] = $ip_permission->getUserId();
        $group['PeeringStatus'] = $ip_permission->getPeeringStatus();
        $group['VpcId'] = $ip_permission->getVpcId();
        $group['VpcPeeringConnectionId'] = $ip_permission->getPeeringConnectionId();

        $permission['UserIdGroupPairs'][] = $group;
      }
    }
    return $permission;
  }

  /**
   * Verify that the revoke was successful.
   *
   * Since OpenStack REST does not return any error codes in the revoke API
   * calls, the only way to verify is to count the permissions array from the
   * current entity, and the entity that is newly updated from the
   * updateSecurityGroups API call.
   *
   * @param string $type
   *   The type of permission.
   * @param \Drupal\cloud\Entity\CloudContentEntityBase $entity
   *   The SecurityGroup.
   *
   * @return bool
   *   True or false.
   */
  private function validateRevoke(string $type, CloudContentEntityBase $entity): bool {
    /** @var \Drupal\openstack\Entity\OpenStackSecurityGroup $updated_group */
    $updated_group = $this->entityTypeManager
      ->getStorage($entity->getEntityTypeId())
      ->load($entity->id());

    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $entity */
    return $type === 'outbound_permission'
      ? $entity->getOutboundPermission()->count() === $updated_group->getOutboundPermission()->count()
      : $entity->getIpPermission()->count() === $updated_group->getIpPermission()->count();
  }

}
